# run with : python3 -m unittest TestFreeShapeModel

import unittest
import os
import numpy as numpy
from astropy import units
import matplotlib.pyplot as plt
import warnings

from StdTests import stdModeltest

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestFreeShapeModel( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )



    def test1( self ):
        x  = numpy.linspace( 0, 10, 101, dtype=float )
        x += 0.01           ## avoid awkward non-continuities
        print( "****** FREESHAPE test 1 ********************" )
        m = FreeShapeModel( 10 )
        self.assertTrue( m.npbase == 10 )
        p = numpy.asarray( [1.2,0.2,1.0,0.5,0.3,1.2,0.2,1.0,0.5,0.3] )

        stdModeltest( m, p, x=x, plot=self.doplot )

    def test2( self ):
        x  = numpy.linspace( 0, 10, 101, dtype=float )
        x += 0.01           ## avoid awkward non-continuities
        print( "****** FREESHAPE test 2 ********************" )
        m = FreeShapeModel( 10, nconvolve=3, center=0.0 )
        p = numpy.asarray( [1.2,0.2,1.0,0.5,0.3,1.2,0.2,1.0,0.5,0.3] )

        stdModeltest( m, p, x=x, plot=self.doplot )

    def xtest3( self ):
        x  = numpy.arange( 21, dtype=float )
        print( "****** FREESHAPE test 3 ********************" )
        m = FreeShapeModel( 8, convolve=4 )
        print( m.npbase )
        self.assertTrue( m.npbase == 8 )
        p = numpy.asarray( [1.2,0.2,1.1,0.5,0.3,0.1,0.4,0.2], dtype=float )

        stdModeltest( m, p, x=x, plot=self.doplot )

    def test4( self ) :
        print( "****** FREESHAPE test 4 ********************" )
        if not self.doplot :
            return

        x  = numpy.linspace( 0, 10, 101, dtype=float )
        for k in range( 7 ) :
            m = FreeShapeModel( 10, nconvolve=k, center=0.0 )
            self.assertTrue( m.npbase == 10 )
            p = numpy.asarray( [1.2,0.9,0.7,0.5,0.3,0.3,0.6,1.0,0.9,0.9] )
            plt.plot( x, m.result( x, p ) )

        plt.show()

    def test5( self ) :
        print( "****** FREESHAPE test 5 ********************" )
        if not self.doplot :
            return

        x  = numpy.linspace( 0, 10, 101, dtype=float )
        p = numpy.asarray( [1.2,0.9,0.7,0.5,0.3,0.3,0.6,1.0,0.9,0.9] )

        m0 = FreeShapeModel( 10, nconvolve=0, center=0.0 )
        m1 = FreeShapeModel( 10, nconvolve=0, center=0.5 )
        plt.plot( x, m0.result( x, p ) )
        plt.plot( x, m1.result( x, p ) )

        m0 = FreeShapeModel( 10, nconvolve=3, center=0.0 )
        m1 = FreeShapeModel( 10, nconvolve=3, center=0.5 )
        plt.plot( x, m0.result( x, p ) )
        plt.plot( x, m1.result( x, p ) )

        plt.show()

    def xtest9( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "****** FREESHAPE test 9 ********************" )
        m = FreeShapeModel( 15, pixperbin=3, convolve=2, xlo=-1.1, xhi=1.1 )
        print( m.npbase )
        self.assertTrue( m.npbase == 6 )
        p = numpy.asarray( [1.2,0.2,1.0,0.5,0.3,0.1,0.4], dtype=float )

        stdModeltest( m, p, plot=self.doplot )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestFreeShapeModels.__class__ )

if __name__ == '__main__':
    unittest.main( )





import numpy as numpy

from .Modifiable import Modifiable
from .Engine import Engine
from . import Tools
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL"
__version__ = "2.6.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2020 Do Kester

class StructureEngine( Engine ):
    """
    The StructureEngine varies the internal structure of the model.

    Only for Models that are Modifiable.

    The member is kept when the logLikelihood > lowLhood.

    Attributes
    ----------
    None of its own

    Attributes from Engine
    ----------------------
    walkers, errdis, slow, maxtrials, rng, report, unitRange, unitMin, verbose

    Author       Do Kester.

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, slow=None, copy=None, seed=23455, verbose=0 ) :
        """
        Constructor.

        Parameters
        ----------
        walkers : list of Walker
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        slow : None or int > 0
            Run this engine every slow-th iteration. None for all.
        copy : StructureEngine
            to be copied
        seed : int
            for random number generator

        """
        super( ).__init__( walkers, errdis, slow=slow, copy=copy,
                    seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return StructureEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "StructureEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """
        Execute the engine by changing a component.

        Parameters
        ----------
        kw : int
            index of walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place or append in walkerlist
        iteration : int
            iteration number

        Returns
        -------
        int : the number of successfull moves

        """
        self.reportCall()

        t = 0
        k = 0
        perm = self.rng.permutation( self.walkers[kw].problem.model.ncomp - 2 )     ## TBC
        for p in perm :
            update = len( self.walkers ) if append else kw
            dt = self.executeOnce( kw, lowLhood, update, location=p+1 )
            if dt > 0 :
                kw = update
            t += dt

        return t

    def executeOnce( self, wlkrid, lowLhood, update, location=None ) :
        """
        One execution call.
        """

        walker = self.walkers[wlkrid]
        cwalker = walker.copy()                  ## work on local copy.
        problem = cwalker.problem.copy()
        ptry = cwalker.allpars.copy()
        ftry = cwalker.fitIndex.copy()

        if self.verbose > 4 :
            print( "SEN0  ", walker.id, walker.parent, len( ptry ), len( ftry ), walker.logL )
            print( "      ", fmt( problem.model.knots, max=None ) )

        off = 0
        model = problem.model
        while model is not None and not isinstance( model, Modifiable ) :
            off += model.npbase
            model = model._next

        if self.verbose > 4 :
            nc = model.ncomp
            np = model.npbase
            print( "SEN1  ", cwalker.id, nc, np, len( ptry ), len( ftry ) )

        if not model.vary( location=location, rng=self.rng ):
            self.reportReject()
            return 0

        problem.model = model._head
        Ltry = self.errdis.logLikelihood( problem, ptry )

        if self.verbose > 4 :
            print( "SEN3  ", fmt( ptry, max=None ), Ltry, lowLhood )
            print( "      ", fmt( problem.model.knots, max=None ) )

        if Ltry >= lowLhood:
            self.reportSuccess()
            self.setWalker( update, problem, ptry, Ltry, fitIndex=ftry )
            wlkr = self.walkers[update]

            wlkr.check( nhyp=self.errdis.nphypar )

            return 1

        self.reportFailed()
        return 0





# run with : python3 -m unittest TestTools

import unittest
import numpy as numpy
import math
from datetime import date

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2006-2015 Do Kester (JAVA CODE)

class ATestClass( object ) :
    def __init__( self, copy=None ) :
        super( ATestClass, self ).__init__( )
        self.setatt( "n1", 4 )
        self.setatt( "n2", 4, type=int )
        self.setatt( "p1", 4.0, type=float, isnone=True )
        self.setatt( "p2", None, type=float, isnone=True )
        self.setatt( "ln", 4, type=int, islist=True )
        self.setatt( "lp", 4, type=float, islist=True )

        try :
            self.setatt( "ep", 4.78, type=int, islist=True )
            assert( False )
        except TypeError :
            assert( True )

    def setatt( self, name, value, type=None, islist=None, isnone=False ) :
        Tools.setAttribute( self, name, value, type=type, islist=islist, isnone=isnone )


class TestTools( unittest.TestCase ) :
    """
    Test harness for Fitter class.

    @author Do Kester

    """
    def testATestClass( self ) :
        atc = ATestClass()

        Tools.printclass( atc )

    def testIsBetween( self ) :
        self.assertTrue( Tools.isBetween( 1, 2, 3 ) )
        self.assertTrue( Tools.isBetween( 3, 2, 1 ) )

        self.assertFalse( Tools.isBetween( 2, 1, 3 ) )
        self.assertFalse( Tools.isBetween( 2, 3, 1 ) )


        self.assertTrue( Tools.isBetween( 2, 2, 3 ) )
        self.assertTrue( Tools.isBetween( 1, 2, 2 ) )
        self.assertTrue( Tools.isBetween( 2, 2, 1 ) )
        self.assertTrue( Tools.isBetween( 3, 2, 2 ) )
        self.assertTrue( Tools.isBetween( 2, 2, 2 ) )


    def testBaseModel( self ) :
        atc = BaseModel( 4 )
        atc.posIndex = [1,2]
        print( atc.posIndex )

#        atc.posIndex = 1.2

        try :
            atc.posIndex = 1.2
            print( atc.posIndex )
            self.assertTrue( False )
        except TypeError :
            self.assertTrue( True )

        try :
            atc.PosIndex = 1
            self.assertTrue( False )
        except AttributeError :
            self.assertTrue( True )

        try :
            atc.posIndex = None
            self.assertTrue( False )
        except TypeError :
            self.assertTrue( True )


    def testtoday( self ):
        print( date.today() )

    def testMakeNext( self ):
        print( "===== makeNext ===========================" )
        x = 2
        xgen = Tools.makeNext( x, 0 )
        for k in range( 10 ) :
            nx = next( xgen )
            print( k, nx )
            self.assertTrue( nx == 2 )
        xgen.close()

        x = [2,3,4,5]
        xgen = Tools.makeNext( x, 0 )
        for k in range( 10 ) :
            nx = next( xgen )
            print( k, nx )
            self.assertTrue( ( nx == (k+2) and k < 4 ) or nx == 5 )
        xgen.close()

        x = [2,3,4,5,6,7,8,9,10,11]
        xgen = Tools.makeNext( x, 0 )
        for k in range( 10 ) :
            nx = next( xgen )
            print( k, nx )
            self.assertTrue( nx == (k+2) )

        x = [2,3,4,5]
        xgen = Tools.makeNext( x, 2 )
        for k in range( 10 ) :
            nx = next( xgen )
            print( k, nx )
            self.assertTrue( ( nx == (k+4) and k < 6 ) or nx == 5 )
        xgen.close()

    def testLength( self ) :
        print( "===== length ================================" )
        x = 3.0
        self.assertTrue( Tools.length( x ) == 1 )
        x = [2,3,4,5]
        self.assertTrue( Tools.length( x ) == 4 )
        x = numpy.asarray( [2,3,4,5] )
        self.assertTrue( Tools.length( x ) == 4 )
        x.resize( 2, 2 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( Tools.length( x ) == 2 )
        x = [2,3,4,5]
        x = numpy.asarray( [x] )
        print( x, x.shape, x.__class__ )
        self.assertTrue( Tools.length( x ) == 1 )

    def testToArray( self ) :
        print( "===== toArray ================================" )
        x = Tools.toArray( 3 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 1 )
        x = Tools.toArray( 3.0 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 1 )
        x = Tools.toArray( [3,4,5,6] )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 1 )
        x = Tools.toArray( [3,4.5,5,6], ndim=1 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 1 )
        x = Tools.toArray( [3,4.5], ndim=2 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 2 )
        x = Tools.toArray( [3,4.5,5], ndim=3 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 2 )
        x = Tools.toArray( [[3,4.5],[6,5]], ndim=2 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 2 )

    def testAverage( self ) :
        print( "===== Average ================================" )
        xx = numpy.asarray( [10,20,30], dtype=float )
        aver = numpy.average( xx )
        stdv = numpy.std( xx )

        ax1, sx1 = Tools.average( xx )
        print( ax1, sx1 )
        self.assertTrue( ax1 == aver )
        self.assertTrue( sx1 == stdv )

        ax2, sx2 = Tools.average( xx, circular=[0,360] )
        print( ax2, sx2 )
        self.assertAlmostEqual( ax2, aver )
        self.assertAlmostEqual( sx2, stdv )

        ax3, sx3 = Tools.average( xx, circular=[-180,180] )
        print( ax3, sx3 )
        self.assertAlmostEqual( ax3, aver )
        self.assertAlmostEqual( sx3, stdv )

        xx -= 15
        xx[0] += 360
        aver -= 15

        ax1, sx1 = Tools.average( xx )
        print( ax1, sx1 )
        self.assertFalse( ax1 == aver )
        self.assertFalse( sx1 == stdv )

        ax2, sx2 = Tools.average( xx, circular=[0,360] )
        print( ax2, sx2 )
        self.assertAlmostEqual( ax2, aver )
        self.assertAlmostEqual( sx2, stdv )

        ax3, sx3 = Tools.average( xx, circular=[-180,180] )
        print( ax3, sx3 )
        self.assertAlmostEqual( ax3, aver )
        self.assertAlmostEqual( sx3, stdv )

        wgt = numpy.array( [0.7, 1.4, 0.7] )
        stdv = math.sqrt( numpy.average( [100, 0, 100], weights=wgt ) )
        print( aver, stdv )

        ax1, sx1 = Tools.average( xx )
        print( ax1, sx1 )
        self.assertFalse( ax1 == aver )
        self.assertFalse( sx1 == stdv )

        ax2, sx2 = Tools.average( xx, circular=[0,360], weights=wgt )
        print( ax2, sx2 )
        self.assertAlmostEqual( ax2, aver )
        self.assertAlmostEqual( sx2, stdv )

        ax3, sx3 = Tools.average( xx, circular=[-180,180], weights=wgt )
        print( ax3, sx3 )
        self.assertAlmostEqual( ax3, aver )
        self.assertAlmostEqual( sx3, stdv )

    def testNicenumber( self ) :
        self.oneNiceNumbertest( 0.00634, 0.006 )
        self.oneNiceNumbertest( -0.013, -0.01 )
        self.oneNiceNumbertest( 0.13, 0.1 )
        self.oneNiceNumbertest( 0, 0 )
        self.oneNiceNumbertest( 1.3, 1 )
        self.oneNiceNumbertest( 31.3, 30 )
        self.oneNiceNumbertest( 801.3, 800 )

        self.oneNiceNumbertest( -1.3, -1 )
        self.oneNiceNumbertest( -5.3, -5 )
        self.oneNiceNumbertest( -1001.3, -1000 )


    def oneNiceNumbertest( self, x, n ) :
        k = Tools.nicenumber( x )
        print( "nicenumber of %f is %f (%f)" % ( x, k, n ) )
        self.assertTrue( k == n )




    @classmethod
    def suite( cls ):
        return ConfiguredTestCase.suite( PriorTest.__class__ )

if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest TestSoftMaxModel.Test.test1

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt
import warnings

from numpy.testing import assert_array_almost_equal as assertAAE
from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for SoftMax Model

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )
        self.dofull = ( "DOFULL" in os.environ and os.environ["DOFULL"] == "1" )


    def test1( self ):

        print( "****** SOFTMAX test 1 ***************" )

        m = SoftMaxModel( ndim=5, ndout=2, offset=False )
        self.assertTrue( m.npbase == 10 )
        self.assertFalse( m.offset )
        self.assertTrue( m.in2out == 10 )
        self.assertTrue( m.ndim == 5 )
        self.assertTrue( m.ndout == 2 )
        cm = m.copy()
        self.assertTrue( cm.npbase == 10 )
        self.assertFalse( cm.offset )
        self.assertTrue( cm.in2out == 10 )
        self.assertTrue( cm.ndim == 5 )
        self.assertTrue( cm.ndout == 2 )

        m = SoftMaxModel( ndim=7, ndout=3, offset=True )
        self.assertTrue( m.npbase == 24 )
        self.assertTrue( m.offset )
        self.assertTrue( m.in2out == 21 )
        self.assertTrue( m.ndim == 7 )
        self.assertTrue( m.ndout == 3 )
        cm = m.copy()
        self.assertTrue( cm.npbase == 24 )
        self.assertTrue( cm.offset )
        self.assertTrue( cm.in2out == 21 )
        self.assertTrue( cm.ndim == 7 )
        self.assertTrue( cm.ndout == 3 )

    def test1a( self ) :
        model = SoftMaxModel( ndim=1, ndout=1, offset=True )

        print( model.npars )
        p = numpy.asarray( [1.0, 1.0] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        stdModeltest( model, p, x=x, plot=self.doplot )
 
    def test1b( self ) :
        model = SoftMaxModel( ndim=1, ndout=2, offset=True, normed=False )

        print( model.npars )
        p = numpy.asarray( [1.0, 1.0, 2.0, 2.0] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        stdModeltest( model, p, x=x, plot=self.doplot )
 


    def makeData( self, nin, npt, ndout ) :
        nn = npt // ndout
        xc = numpy.asarray( [k for k in range( ndout )] * nn, dtype=float )
        numpy.random.seed( 12345 )
        x = xc + 0.3 * numpy.random.randn( nin*npt ).reshape( nin, npt )
        return ( x.transpose(), xc )


    def test2a( self ) :
        nin = 5
        npt = 9
        ndout = 3

        x, y = self.makeData( nin, npt, ndout )

        print( "****** SOFTMAX test 2a False **************" )

        m = SoftMaxModel( ndim=nin, ndout=ndout, normed=False )

        p = numpy.linspace( -1.0, 1.0, m.npars )
        print( " In %d  ND %d  NP %d  Out %d" % (nin, npt, m.npars, ndout) )

        print( "param  ", p )

        stdModeltest( m, p, x=x, plot=self.doplot )


    def test2b( self ) :
        nin = 5
        npt = 9
        ndout = 3

        x, y = self.makeData( nin, npt, ndout )

        print( "****** SOFTMAX test 2b True ***************" )

        m = SoftMaxModel( ndim=nin, ndout=ndout, normed=True )

        p = numpy.linspace( -1.0, 1.0, m.npars )
        print( " In %d  ND %d  NP %d  Out %d" % (nin, npt, m.npars, ndout) )

        print( "param  ", p )

        stdModeltest( m, p, x=x, plot=self.doplot )



    def test3( self ) :
        nin = 5
        npt = 9
        ndout = 3
        x, y = self.makeData( nin, npt, ndout )

        print( fma( x ) )

        print( "****** SOFTMAX test 2 ***************" )
        m = SoftMaxModel( ndim=nin, ndout=ndout, offset=True )

        p = numpy.linspace( -1.0, 1.0, m.npars )
        print( "param  ", p )

        stdModeltest( m, p, x=x, plot=self.doplot )

    def xxxdtest( self, x, m, p ) :

        f = m.result( x, p )
        print( "Result   ", f.shape )
        print( fma( f ) )

        if self.doplot :
            for k in range( x.shape[1] ) :
                plt.plot( x[:,k], 'k-' )
            for n in range( f.shape[1] ) :
                plt.plot( f[:,n], 'r-' )
            plt.show()


        dfdx = m.derivative( x, p )
        print( "DfDx     ", dfdx.shape )

        for k in range( x.shape[1] ) :
            xp = x.copy()
            xm = x.copy()

            xp[:,k] += 0.00001
            xm[:,k] -= 0.00001

            yp = m.result( xp, p )
            ym = m.result( xm, p )

            numx = ( yp - ym ) / 0.00002
            numx = numx.transpose()

            print( k )
            print( "dx0   ", fma( dfdx[:,:,k], indent=7 ) )
            print( "nm0   ", fma( numx, indent=7 ) )
            assertAAE( dfdx[:,:,k], numx )

        part = m.basePartial( x, p )
        for pk in part :
            print( "Partial  ", pk.shape, id( pk ) )
            print( fmt( pk, tail=6 ) )

        for k in range( m.npars ) :
            print( "parameter ", k, "  ", m.parNames[k] )
            pp = p.copy()
            pp[k] += 0.0001
            pm = p.copy()
            pm[k] -= 0.0001
            yp = m.result( x, pp )
            ym = m.result( x, pm )

            num = ( yp - ym ) / 0.0002

            mm = k // m.ndim
            if mm >= m.ndout : 
                mm = k - m.in2out

            print( "In ", k, "  Out ", mm )
            print( "part  ", fmt( part[mm][:,k], tail=3 ) )
            print( "nm0   ", fmt( num[:,mm], tail=3 ) )
            assertAAE( part[mm][:,k], num[:,mm], 4 )



    def test4a( self ) :
        self.dotest4( 1, 3, 2, True, True )

    def test4b( self ) :
        self.dotest4( 2, 4, 3, False, False )

    def test4c( self ) :
        self.dotest4( 3, 3, 2, True, False )

    def test4d( self ) :
        self.dotest4( 2, 3, 2, False, True )


    def dotest4( self, nin, npt, ndout, offset, normed ) :

        m = SoftMaxModel( ndim=nin, ndout=ndout, offset=offset, normed=normed )

        print( "****** SOFTMAX test 4 **** ", m.ndim, m.in2out, m.ndout, "***********" )
        print( "******** offset=", m.offset, "  normed=", m.normed )

#        x, y = self.makeData( nin, npt, ndout )
        
#        x = numpy.array( [[0,0,0,0,0],[1,2,3,4,5]], dtype=float )
        numpy.random.seed( 12345 )

        x = numpy.random.randn( npt, nin )
        y = numpy.random.randint( 0, high=ndout, size=npt )

        pars = numpy.random.randn( m.npars )
        print( "Pars  ", fma( pars ) )
        par = pars[:m.in2out]

        print( "X     ", fma( x, indent=7 ) )
        print( "Y     ", fma( y ) )
        print( "Res   ", fma( m.result( x, pars ), indent=7 ) )

        mprob = MultipleOutputProblem( m, x, y )
        prob = ClassicProblem( m, x, y )

        ppart = prob.partial( pars )
        print( "problem partial" )
        print( fmt( ppart, tail=6 ) ) 

        npart = m.strictNumericPartial( x, pars )
        print( "problem numpart" )
        print( fmt( npart, tail=6 ) ) 

        assertAAE( ppart, npart )


        bed = BernoulliErrorDistribution()

        logLdata = bed.logLdata( mprob, pars )
        print( "LLdata ", fma( logLdata ) )
        logLdata = bed.logLdata( prob, pars )
        print( "LLdata ", fma( logLdata ) )

        logL = bed.logLikelihood( prob, pars )
        print( "LogL   ", fma( logL ) )
        altL = bed.logLikelihood_alt( prob, pars )
        print( "altL   ", fma( altL ) )


        fi = [k for k in range( m.npars )]
        dL = bed.partialLogL_alt( prob, pars, fi )

        nL = numpy.zeros_like( dL )

        pg = bed.nextPartialData( prob, pars, fi )
        np = len( fi )
        pL = numpy.zeros( np, dtype=float )


        for k in range( m.npars ) :
            pp = pars.copy()
            pp[k] += 0.001
            pm = pars.copy()
            pm[k] -= 0.001
            Lp = bed.logLikelihood( prob, pp )
            Lm = bed.logLikelihood( prob, pm )

            nL[k] = ( Lp - Lm ) / 0.002

            pL[k] = numpy.sum( next( pg ) )

            print( fmt(k), fmt(pars[k]), fmt(dL[k]), fmt(nL[k]), fmt(pL[k]) )


        print( fma( dL ) )
        print( fma( nL ) )
        print( fma( pL ) )
        print( fma( bed.numPartialLogL( prob, pars, fi ) ) )

        assertAAE( dL, pL )



    def test5( self ) :
        nin = 5
        npt = 9
        ndout = 3

        x, y = self.makeData( nin, npt, ndout )

        print( "****** SOFTMAX test 5 ***************" )
        m = SoftMaxModel( ndim=nin, ndout=ndout, offset=True )
        m.setLimits( -10, 10 )

        problem = ClassicProblem( model=m, xdata=x, ydata=y )

        erdis = "bernoulli"
        ns = NestedSampler( problem=problem, distribution=erdis, verbose=2 )

        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10

        loge = ns.sample()

        sl = ns.samples
        pars = ns.parameters

        print( fma( y ) )
        print( fma( m.result( x, pars ).T ) )

        self.plotresults( x, m, y, pars )

    def test6a( self ) :
        npt = 30
        ninp = [1, 1, 8, 8]
        nout = [1, 1, 1, 3]
        nhid = [1, 5, 5, 5]

        for (ni,no,nh) in zip( ninp, nout, nhid ):
            print( ni, no, nh, npt )
            problem = self.makeModel( ni, no, nh, npt )
            

    def makeModel( self, nin, ndout, ndhid, npt ) :
        x, y = self.makeData( nin, npt, ndout )

        m1 = SoftMaxModel( ndim=nin, ndout=ndhid, offset=True )
        m1.setLimits( -10, 10 )

        p1 = numpy.random.random( m1.npars ) * 20 - 10
        print( m1.npars, fmt( m1.result( x, p1 ) ) )

        m2 = SoftMaxModel( ndim=ndhid, ndout=ndout, offset=True )
        m2.setLimits( -10, 10 )

        m = m1 | m2

        pp = numpy.random.random( m.npars ) * 20 - 10
        print( m.npars, fmt( m.result( x, pp ) ) )


        problem = ClassicProblem( model=m, xdata=x, ydata=y )
        return problem


    def test6( self ) :
        print( "****** SOFTMAX test 6 ***************" )

        nin = 8
        npt = 30
        ndout = 3
        nhid = 5

        problem = self.makeModel( nin, ndout, nhid, npt )
        x, y = self.makeData( nin, npt, ndout )

        erdis = "bernoulli"
        ns = NestedSampler( problem=problem, distribution=erdis, verbose=2 )

        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10

        loge = ns.sample()

        sl = ns.samples
        pars = ns.parameters

        print( fma( y ) )
        m = problem.model
        print( fma( m.result( x, pars ).T ) )


        self.plotresults( x, m, y, pars )


    def makeMap( self, npt, ndout ) :
        nn = npt // ndout
        t = []
        x = []

        numpy.random.seed( 12345 )
#        cen4 = [[0.5,0.5],[0,0],[0,1],[1,0],[1,1]]
        cen4 = [[0,1],[1,1],[0,0]]
        for n in range( ndout ) :
            t += [n] * nn

            cen = numpy.random.random( 2 )
            rot = numpy.random.random( 4 )
            rot[1:3] = ( rot[1] + rot[2] ) / 2
            rot[1:3] = 0.0
            rot[0] = 0.1
            rot[3] = 0.1

            cen = cen4[n]
            rot = rot.reshape( 2, 2 )
            xx = numpy.random.randn( 2 * nn ).reshape( nn, 2 )
            xx = numpy.inner( xx, rot )
            x = ( xx + cen ) if n == 0 else numpy.append( x, xx + cen, axis=0 )

        x = numpy.where( x > 1, 2 - x, x )
        x = numpy.where( x < 0, -x, x )

#        print( fma( x.T ) )
#        print( fma( t ) )
 
        return ( x, numpy.array( t ) )

    def test7( self ) :
        nin = 2
        npt = 600
        ndout = 3

        x, y = self.makeMap( npt, ndout )

        print( "****** SOFTMAX test 7 ***************" )
        m = SoftMaxModel( ndim=nin, ndout=ndout, offset=True )
        m.setLimits( -10, 10 )

        problem = ClassicProblem( model=m, xdata=x, ydata=y )

        erdis = "bernoulli"
        ns = NestedSampler( problem=problem, distribution=erdis, verbose=2 )

        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10

        loge = ns.sample()

        sl = ns.samples
        pars = ns.parameters

#        print( fma( y ) )
#        print( fma( m.result( x, pars ).T ) )

        if not self.doplot : return

#        cc = ['r.', 'g.', 'b.', 'c.', 'm.', 'y.', 'o.']
        cc = ['black', 'red', 'green', 'blue', 'cyan', 'magenta', 'yellow', 'orange']
        for k in range( ndout ) :
            q = numpy.where( y == k )
            plt.plot( x[q,1], x[q,0], color=cc[k], marker='.' )

        map = numpy.zeros( (21,21), dtype=float )
        ias = ImageAssistant()
        xx = ias.getIndices( map ) / 20.0

        yy = m.result( xx, pars )
        ax = numpy.linspace( 0, 1, 21 )
        v = numpy.linspace( 0.05, 0.95, 3 )
        v = [0.15, 0.55, 0.95]
        for k in range( ndout ) :
            map = ias.resizeData( yy[:,k], shape=[21,21] )
            plt.contour( ax, ax, map, v, colors=cc[k] )


        plt.show()        






    def plotresults( self, x, model, y, pars ) :

        if not self.doplot : return

        t = numpy.arange( len( y ), dtype=float )
#        print( fmt( t, max=10 ) )
        tt = numpy.append( t, t + 0.99 )
        ii = numpy.argsort( tt )
#        print( fmt( tt[ii], max=10 ) )
        yh = numpy.append( y, y )
#        print( fmt( y, max=10 ) )
#        print( fmt( yh[ii], max=10 ) )

        plt.plot( tt[ii], yh[ii] + 1, 'k-' )
	
        yfit = model.result( x, pars )
        dt = 1 / model.lastndout 
        t0 = 0.5 * dt
        cc = ['r.', 'g.', 'b.']
        for k in range( model.lastndout ) :
            plt.plot( t + 0.5, y + yfit[:,k], cc[k] )
#            t0 += dt

        plt.show()






    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestSoftMaxModel.__class__ )

if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest Test2dModels

import unittest
import os
import numpy as numpy
from astropy import units
import matplotlib.pyplot as plt
import warnings

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test2dFixedModels( unittest.TestCase ):
    """
    Test harness for 2d-Models

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def testPolySineAmpModel( self ):
        x  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******POLYSINEAMP********************" )

        self.assertRaises( AttributeError, PolySineAmpModel, 2, 1.3, fixed={3:1.1} )

    def testPolySurfaceModel( self ):
        x  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******POLYSURFACE********************" )
        m = PolySurfaceModel( 2, fixed={2:0, 5:0} )
        self.assertTrue( m.npmax == 6 )
        self.assertTrue( m.npchain == 4 )
        self.assertTrue( m.npbase == 4 )
        p = [-1.1, 0.04, 1.2, -0.03]
        self.stdModeltest( m, p, plot=self.doplot )

    def testEtalonDriftModel( self ):
        x  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******ETALON DRIFT********************" )
        m = EtalonDriftModel( fixed={0:2, 1:10} )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [0.04, 1.2, -0.5]
        self.stdModeltest( m, p, plot=self.doplot )

    def testKernel2dModel1( self ):
        x  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******KERNEL2D CIRCULAR********************" )
        m = Kernel2dModel( fixed={0:1})
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [0.5, 0.04, 1.2]
        self.stdModeltest( m, p, plot=self.doplot )

    def testKernel2dModel2( self ):
        x  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******KERNEL2D ELLIPTIC********************" )
        m = Kernel2dModel( shape=2, fixed={0:1} )
        self.assertTrue( m.npmax == 5 )
        self.assertTrue( m.npchain == 4 )
        self.assertTrue( m.npbase == 4 )
        p = [0.5, 0.04, 1.2, 1.4]
        self.stdModeltest( m, p, plot=self.doplot )

    def testKernel2dModel3( self ):
        x  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******KERNEL2D ROTATED********************" )
        m = Kernel2dModel( shape=3, fixed={0:1} )
        self.assertTrue( m.npchain == 5 )
        self.assertTrue( m.npbase == 5 )
        p = [0.5, 0.04, 1.2, 1.3, 1.1]
        self.stdModeltest( m, p, plot=self.doplot )

    def testProductModel( self ):
        rng = numpy.random.RandomState( 1234 )
        x = rng.rand( 100, 2 )
        x[:,0] *= 3
        x[:,1] *= 2

        xy  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******2D PRODUCT*****************" )
        gm = GaussModel()
        pm = PolynomialModel( 2 )
        self.assertRaises( AttributeError, ProductModel, [gm,pm], fixed={3:1.1} )


    def testSurfaceSplinesModel( self ):
        rng = numpy.random.RandomState( 1234 )
        x = rng.rand( 100, 2 )
        x[:,0] *= 3
        x[:,1] *= 2

        xy  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******SURFACESPLINES*****************" )
        xk = numpy.arange( 4, dtype=float )
        yk = numpy.arange( 3, dtype=float )
        self.assertRaises( AttributeError, SurfaceSplinesModel, [xk,yk], fixed={3:1.1} )

    """
    def testFreeShape2dModel( self, plot=False ):
        rng = numpy.random.RandomState( 1234 )
        x = rng.rand( 100, 2 )
        x[:,0] *= 3
        x[:,1] *= 2

        xy  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        print( "******FREESHAPW2DSPLINES*****************" )
        xk = numpy.arange( 4, dtype=float )
        yk = numpy.arange( 3, dtype=float )
        self.assertRaises( AttributeError, FreeShape2dModel, (12,13), fixed={3:1.1} )
    """


    def stdModeltest( self, model, par, xdata=None, plot=None, warn=[] ):
        print( "***StdModelTest***************" )
        if xdata is None :
            x  = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                    [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        else :
            x = xdata

        print( model )
        if "nopart" in warn :
            self.assertWarns( UserWarning, model.basePartial, x, par )
            print( model.shortName() + ": Further no-partial warnings ignored." )
            warnings.simplefilter( "ignore" )

        numpy.set_printoptions( precision=3, suppress=True )
        print( model.partial( x, par ) )

        model.xUnit = [units.m, units.s]
        model.yUnit = units.kg
        for k in range( model.getNumberOfParameters() ):
            print( "%d  %-12s  %-12s"%(k, model.getParameterName( k ), model.getParameterUnit( k ) ) )

        print( "Integral   ", model.getIntegralUnit( ) )
        self.assertTrue( model.testPartial( x, par ) == 0 )

        part = model.partial( x, par )
        nump = model.numPartial( x, par )
        k = 0
        for (pp,nn) in zip( part.flatten(), nump.flatten() ) :
            self.assertAlmostEqual( pp, nn, 3 )
            k += 1

        mc = model.copy( )
        Tools.printclass( model )
        Tools.printclass( mc )
        for k in range( min( mc.npchain, 10 ) ):
            print( "%d  %-12s  %-12s"%(k, mc.getParameterName( k ), mc.getParameterUnit( k ) ) )
        mc.parameters = par
        model.parameters = par

        for (k,xk,r1,r2) in zip( range( 11 ), x, model.result( x ), mc.result( x ) ) :
            print( "%3d %8.3f,%8.3f  %10.3f %10.3f" % ( k, xk[0], xk[1], r1, r2 ) )
            self.assertEqual( r1, r2 )
        self.assertTrue( mc.testPartial( x, par ) == 0 )

#        if plot :
#            self.plotModel( model, par )

    def TBDplotModel( self, model, par ) :
        xx = numpy.linspace( -1, +1, 1001 )
        plt.plot( xx, model.result( xx, par ), '-', linewidth=2 )
        x2 = numpy.linspace( -1, +1, 11 )
        dy = model.derivative( x2, par )
        yy = model.result( x2, par )
        for k in range( 11 ) :
            x3 = numpy.asarray( [-0.05, +0.05] )
            y3 = x3 * dy[k] + yy[k]
            plt.plot( x2[k] + x3, y3, 'r-' )

        plt.show()


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestModels.__class__ )

if __name__ == '__main__':
    unittest.main( )





#run with: python3 -m unittest TestPipeModel

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

#from BayesicFitting import PolynomialModel, SineModel
#from Model import Model

from StdTests import stdModeltest

from BayesicFitting import formatter as fmt
from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def testPipe1( self ) :
        print( "=== testPipe1:  1 ==> 1 ==> 1  pipe_0 and pipe_4 ==============" )

        model = PolynomialModel( 2 )
        model |= SineModel()
        self.pipetest( model, 0, 4 )

        print( model )
        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 1.0] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        self.shapetest( model, x, p )

        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe2( self ) :
        print( "=== testPipe2:  pipe_0 and pipe_4 ==============" )
        model = PolynomialModel( 2 )
        model |= SineModel()
        model *= ExpModel()
        print( model )
        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 1.0, 1.3, -0.2] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        self.shapetest( model, x, p )

        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe3( self ) :
        print( "=== testPipe3:  1 ==> 2 ==> 1  pipe_2 and pipe_6 ==============" )
        model = StellarOrbitModel( )
        model |= PolySurfaceModel( 3 )
        self.pipetest( model, 2, 6 )

        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 0.0, 0.4, 0.2, 1.0, 0.0, 0.0, 0.4,
                            -0.1, -0.2, 0.3, 0.0, 0.0] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        self.shapetest( model, x, p )
        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe4( self ) :
        print( "=== testPipe4:  1 ==> 1 ==> 2  pipe_3 and pipe_7 ==============" )
        model = PolynomialModel( 1 )
        model |= StellarOrbitModel( )
        self.pipetest( model, 3, 7 )

        p = numpy.asarray( [0.1, 1.1, 0.0, 0.4, 0.2, 1.0, 0.0, 0.0, 0.4] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        self.shapetest( model, x, p )
        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe5( self ) :
        print( "=== testPipe5: 2 ==> 1 ==> 2  pipe_3 and pipe_9 ==============" )
        model = PolySurfaceModel( 1 )
        model |= StellarOrbitModel( )
        self.pipetest( model, 3, 9 )

        p = numpy.asarray( [0.1, 1.1, 0.9, 0.0, 0.4, 0.2, 1.0, 0.0, 0.0, 0.4] )
        x = numpy.linspace( 0, 10, 101*model.ndim, dtype=float ).reshape( 101, model.ndim )

        self.shapetest( model, x, p )
        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe6( self ) :
        print( "=== testPipe6: 2 ==> 1 ==> 1  pipe_0 and pipe_0 ==============" )
        model = PolySurfaceModel( 1 )
        model |= ArctanModel( )
        self.pipetest( model, 0, 0 )

        p = numpy.asarray( [0.1, 1.1, 0.9, 2.0, 0.4, 0.2] )
        x = numpy.linspace( 0, 10, 101*model.ndim, dtype=float ).reshape( 101, model.ndim )

        self.shapetest( model, x, p )
        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe7a( self ) :
        print( "=== testPipe7a: check Raise  ==============" )
        model = SoftMaxModel( ndim=3, ndout=2, normed=False )
        m2 = StellarOrbitModel( )
        self.assertRaises( ValueError, model.appendModel, m2, model.PIP )

    def testPipe7( self ) :
        print( "=== testPipe7: 1 ==> 2 ==> 3  pipe_1 and pipe_5 ==============" )

        nx = 2
        x1 = numpy.zeros( nx, dtype=float ) + 0.03
                
        model1 = StellarOrbitModel()
        p1 = [0.1, 1.0, 2.0, 0.0, 0.0, 0.0, 0.0]
        print( fmt( x1 ) )

        x2 = model1.result( x1, p1 )
        print( fmt( x2 ) )

        model2 = SoftMaxModel( ndim=2, ndout=3, normed=False )

        p2 = numpy.arange( model2.npars, dtype=float ) + 1

        print( fmt( model2.result( x2, p2 ) ) )
        print( fmt( p2, max=None ) )
        print( fmt( model2.partial( x2, p2 ), max=None ) )
        print( fmt( model2.strictNumericPartial( x2, p2 ), max=None ) )
        print( fmt( model2.derivative( x2, p2 ), max=None ) )
        print( fmt( model2.strictNumericDerivative( x2, p2 ), max=None ) )


        model = model1 | model2
        p = numpy.append( p1, p2 )

        printclass( model._next )
        self.pipetest( model, 1, 5 )

        x = x1
        print( fmt( model.result( x, p ) ) )
        print( fmt( p, max=None ) )
        print( fmt( model.partial( x, p ), max=None ) )
        print( fmt( model.strictNumericPartial( x, p ), max=None ) )
        print( fmt( model.derivative( x, p ), max=None ) )
        print( fmt( model.strictNumericDerivative( x, p ), max=None ) )
        

        self.shapetest( model, x, p, nx=nx )

        nx = 11
        x = numpy.zeros( nx, dtype=float ) + 0.03

        stdModeltest( model, p, x=x, plot=self.doplot )


    def testPipe8( self ) :
        print( "=== testPipe8: 3 ==> 2 ==> 1  pipe_2 and pipe_8 ==============" )
        model = SoftMaxModel( ndim=3, ndout=2, normed=False )
        model |= PolySurfaceModel( 1 )
        self.pipetest( model, 2, 8 )

        p = numpy.asarray( [0.1, 1.1, 0.9, 0.0, 0.4, 0.2, 0.1, 1.1, 0.9] )
        x = numpy.linspace( 0, 10, 101*model.ndim, dtype=float ).reshape( 101, model.ndim )

        self.shapetest( model, x, p )
        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe9( self ) :
        print( "=== testPipe9: 3 ==> 2 ==> 4  pipe_1 and pipe_1 ==============" )

        model = SoftMaxModel( ndim=3, ndout=2, normed=False )
        model |= SoftMaxModel( ndim=2, ndout=4, normed=False )
        self.pipetest( model, 1, 1 )

        p = numpy.linspace( -1, 1, model.npars )
        x = numpy.linspace( 0, 10, 101*model.ndim, dtype=float ).reshape( 101, model.ndim )

        self.shapetest( model, x, p )
        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe10( self ) :
        print( "=== testPipe10: 1 ==> 2 ==> 3  pipe_1 and pipe_5 ==============" )
        model = StellarOrbitModel( )
        model |= SoftMaxModel( ndim=2, ndout=3, normed=False )

        printclass( model._next )
        self.pipetest( model, 1, 5 )

        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 0.0, 0.4, 0.2, 1.0, 0.0, 0.0, 0.4, 0.5] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        self.shapetest( model, x, p )
        stdModeltest( model, p, x=x, plot=self.doplot )

    def pipetest( self, model, p, d ) :

        self.assertTrue( "pipe_%d"%p in str( model._next.pipePartial ) )  
        self.assertTrue( "pipe_%d"%d in str( model._next.pipeDeriv ) ) 




    def shapetest( self, model, x, p, nx=101 ) :
        ni = model.ndim
        no = model.lastndout
        np = model.npars

        print( "In %d  Out %d  Xd %d  Np %d" % ( ni, no, nx, np ) )
        xshp = x.shape
        print( "xshape  ", xshp )

        y = model.result( x, p )
        yshp = y.shape
        print( "yshape  ", yshp )
        self.assertTrue( yshp[0] == nx )
        if no > 1 :
            self.assertTrue( yshp[1] == no )

        pp = model.partial( x, p )
        if no > 1 :
            self.assertTrue( len( pp ) == no )
            pshp = pp[0].shape
            print( "pshape  ", len( pp ), pshp )
        else :
            pshp = pp.shape
            print( "pshape  ", pshp )
        self.assertTrue( pshp[0] == nx )
        self.assertTrue( pshp[1] == np )

        dd = model.derivative( x, p )
        if no > 1 and ni > 1 :
            self.assertTrue( len( dd ) == no )
            dshp = dd[0].shape
            print( "dshape  ", len( dd ), dshp )
        else :
            dshp = dd.shape
            print( "dshape  ", dshp )
        self.assertTrue( dshp[0] == nx )
        if no == 1 and ni > 1 :
            self.assertTrue( dshp[1] == ni )
        elif no > 1 and ni == 1 :
            self.assertTrue( dshp[1] == no )


    def suite( cls ):
        return unittest.TestCase.suite( TestPipeModel.__class__ )


if __name__ == '__main__':
    unittest.main()




#run with: python3 -m unittest TestCompoundModel

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 Do Kester

class TestDynamicModel( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )
        self.dofull = ( "DOFULL" in os.environ and os.environ["DOFULL"] == "1" )

    def testDynamic( self ) :
        print( "  Test Dynamic" )
        dyn = Dynamic( dynamic=False )
        self.assertFalse( dyn.isDynamic() )

        dyn = Dynamic(  )
        self.assertTrue( dyn.isDynamic() )

        par = numpy.arange( 8, dtype=float ) * 0.1
        print( "par      ", fmt( par, max=None ) )
        par = dyn.alterParameters( par, 3, 1, 2, value=1.0 )
        print( "par      ", fmt( par, max=None ) )
        par = dyn.alterParameters( par, 4, -1, 2 )
        print( "par      ", fmt( par, max=None ) )
        par = dyn.alterParameters( par, 6, 2, 2, value=[1.0,2.0] )
        print( "par      ", fmt( par, max=None ) )
        par = dyn.alterParameters( par, 8, -1, 2 )
        print( "par      ", fmt( par, max=None ) )
        par = dyn.alterParameters( par, 6, -1, 2 )
        print( "par      ", fmt( par, max=None ) )

        fin = [0,1,2,3,4,5,-2,-1]
        print( "fin      ", fmt( fin, max=None ) )
        fin = dyn.alterFitindex( fin, 3, 1, 2 )
        print( "fin      ", fmt( fin, max=None ) )
        fin = dyn.alterFitindex( fin, 3, -1, 2 )
        print( "fin      ", fmt( fin, max=None ) )
        fin = dyn.alterFitindex( fin, 4, 1, 2 )
        print( "fin      ", fmt( fin, max=None ) )
        fin = dyn.alterFitindex( fin, 5, -1, 2 )
        print( "fin      ", fmt( fin, max=None ) )

        fin = [0,2,3,4,5,-2,-1]
        print( "fin      ", fmt( fin, max=None ) )
        fin = dyn.alterFitindex( fin, 2, 1, 2 )
        print( "fin      ", fmt( fin, max=None ) )
        fin = dyn.alterFitindex( fin, 2, -1, 2 )
        print( "fin      ", fmt( fin, max=None ) )
        fin = dyn.alterFitindex( fin, 3, 1, 2 )
        print( "fin      ", fmt( fin, max=None ) )
        fin = dyn.alterFitindex( fin, 4, -1, 2 )
        print( "fin      ", fmt( fin, max=None ) )


    def test1Model1( self ):
        print( "  Test PolynomialDynamicModel" )
        m = PolynomialDynamicModel( 0 )
        self.growshrink( m, dnp=1 )

    def test1Model2( self ):
        print( "  Test HarmonicDynamicModel" )
        m = HarmonicDynamicModel( 1 )
        self.growshrink( m, dnp=2 )

    def test1Model3( self ):
        print( "  Test RepeatingModel 3" )
        m = GaussModel()
        r = RepeatingModel( 1, m  )
        rc = r.copy()

        self.growshrink( rc, dnp=3 )

    def test1Model4( self ):
        print( "  Test RepeatingModel 4" )
        m = GaussModel()
        r = RepeatingModel( 1, m, same=[2]  )
        self.assertTrue( isinstance( r.growPrior, ExponentialPrior ) )

        rc = r.copy()
        self.assertTrue( isinstance( rc.growPrior, ExponentialPrior ) )

        self.assertTrue( rc.same[0] == 2 )
        self.assertTrue( len( rc.same ) == 1 )

        self.assertTrue( rc.index[0] == 0 )
        self.assertTrue( rc.index[1] == 1 )
        self.assertTrue( len( rc.index ) == 2 )

        Tools.printclass( r )

        self.assertTrue( r.grow() )
        self.assertTrue( r.npbase == 5 )
        self.assertTrue( r.shrink() )
        self.assertTrue( r.npbase == 3 )
        r.shrink()
        self.assertTrue( r.npbase == 0 )
        r.grow()
        self.assertTrue( r.npbase == 3 )


        rc.grow()
        rc.grow()
        self.assertTrue( rc.npbase == 7 )
        self.assertTrue( rc.ncomp == 3 )

    def test1Model5( self ):
        print( "  Test RepeatingModel 6" )
        m = GaussModel()
        r = RepeatingModel( 1, m, same=2, minComp=1, maxComp=3  )

        pars = numpy.asarray( [1.0, -0.4, 0.1] )
        self.assertTrue( isinstance( r.growPrior, UniformPrior ) )
        self.assertTrue( r.grow() )
        self.assertTrue( r.npbase == 5 )

        pars = numpy.asarray( [1.0, -0.4, 0.1, 0.5, 0.0] )
        self.assertTrue( r.grow() )
        self.assertTrue( r.npchain == 7 )

        pars = numpy.asarray( [1.0, -0.4, 0.1, 0.5, 0.0, 0.3, 0.4] )
        self.assertFalse( r.grow() )
        self.assertTrue( r.npbase == 7 )

        pars = numpy.asarray( [1.0, -0.4, 0.1, 0.5, 0.0, 0.3, 0.4] )

        ## location is removed in BirthEngine, normally
        del( r.location )
        stdModeltest( r, pars )


        r.parameters = pars
        print( r.parameters )
        self.assertTrue( r.shrink() )
        self.assertTrue( r.npbase == 5 )
        print( r.parameters )

        self.assertTrue( r.shrink() )
        self.assertTrue( r.npbase == 3 )
        print( r.parameters )

        self.assertFalse( r.shrink() )
        self.assertTrue( r.npbase == 3 )
        print( r.parameters )

    def test1Model6( self ):
        print( "  Test RepeatingModel 6" )
        m = GaussModel()
        r = RepeatingModel( 1, m, same=[2], minComp=1, maxComp=1  )
        self.assertFalse( r.isDynamic() )
        self.assertFalse( r.grow() )
        self.assertTrue( r.npbase == 3 )
        self.assertFalse( r.shrink() )
        self.assertTrue( r.npbase == 3 )

        p = RepeatingModel( 1, m, same=[2], dynamic=False  )
        self.assertFalse( p.isDynamic() )
        self.assertFalse( p.grow() )
        self.assertTrue( p.npbase == 3 )
        self.assertFalse( p.shrink() )
        self.assertTrue( p.npbase == 3 )

    def test1Model7( self ):
        print( "  Test RepeatingModel 7" )
        m = GaussModel()
        r = RepeatingModel( 1, m, minComp=1, maxComp=3  )

        self.assertTrue( isinstance( r.growPrior, UniformPrior ) )
        self.assertTrue( r.grow() )
        self.assertTrue( r.npbase == 6 )
        self.assertTrue( r.grow() )
        self.assertTrue( r.npchain == 9 )

        pars = numpy.asarray( [1.0, 0.1, 0.01, 2.0, 0.2, 0.02, 3.0, 0.3, 0.03] )

        r.parameters = pars
        print( r.parameters )
        self.assertTrue( r.shrink() )
        self.assertTrue( r.npbase == 6 )
        print( r.parameters )
        self.assertTrue( r.shrink() )
        self.assertTrue( r.npbase == 3 )
        print( r.parameters )


    def test1Model8( self ):
        print( "  Test RepeatingModel 8" )
        m = GaussModel()
        r = RepeatingModel( 3, m )

        pars = numpy.arange( 9, dtype=float )
        rng = numpy.random.RandomState( 12345 )
        pars = r.shuffle( pars, 0, 9, rng )
        print( pars )
        p0 = pars[0]
        for k,p in enumerate( pars[1:] ) :
            self.assertTrue( p == p0+1 or k%3 == 2 )
            p0 = p

        pars = numpy.arange( 12, dtype=float )
        rng = numpy.random.RandomState( 12345 )
        pars = r.shuffle( pars, 2, 9, rng )
        print( pars )
        self.assertTrue( pars[0]==0 and pars[1]==1 and pars[-1]==11 )

    def test1Model9( self ):
        print( "  Test RepeatingModel 9" )
        m = GaussModel()
        r = RepeatingModel( 3, m, same=1 )

        pars = numpy.array( [0,1,2,3,5,6,8], dtype=float )
        rng = numpy.random.RandomState( 12345 )
        pars = r.shuffle( pars, 0, 7, rng )
        print( pars )
        self.assertTrue( pars[1] == 1 )
        self.assertTrue( pars[2] == pars[0]+2 )
        self.assertTrue( pars[4] == pars[3]+2 )
        self.assertTrue( pars[6] == pars[5]+2 )


        pars = numpy.array( [0,1,2,3,4,5,7,8,10,11], dtype=float )
        rng = numpy.random.RandomState( 12345 )
        pars = r.shuffle( pars, 2, 7, rng )
        print( pars )
        self.assertTrue( pars[0]==0 and pars[1]==1 and pars[-1]==11 )
        self.assertTrue( pars[3] == 3 )
        self.assertTrue( pars[4] == pars[2]+2 )
        self.assertTrue( pars[6] == pars[5]+2 )
        self.assertTrue( pars[8] == pars[7]+2 )

    def test1Model10( self ):
        print( "  Test SplinesDynamicModel" )
        knots = numpy.linspace( 0, 1, 3 )
        m = SplinesDynamicModel( knots=knots )
        m.setLimits( [-10.0], [10.0] )
        m.parameters = numpy.arange( m.npars, dtype=float )
        rng  = numpy.random.RandomState( 123456 )
        m = self.growshrink( m, dnp=1, npf=4, rng=rng, prnt=True )

        print( "Change Structure" )
        for k in range( 6 ) :
            m.vary( rng=rng )
            self.report( m, prnt=True )

        cm = m.copy()
        Tools.printclass( cm )

    def test1Model11( self ):
        print( "  Test SplinesDynamicModel shrink" )
        xx = numpy.linspace( 0, 10, 501, dtype=float )
        kn1 = numpy.linspace( 0, 10, 16, dtype=float )
        m1 = BasicSplinesModel( knots=kn1 )
        p1 = numpy.zeros( 18, dtype=float )
        p1[7:9] = 1.0

        kn2 = numpy.append( numpy.append( kn1[:7], [5.0] ), kn1[9:] )
        m2 = BasicSplinesModel( knots=kn2 )
        p2 = numpy.zeros( 17, dtype=float )
        p2[7] = 1.4
        print( fmt( kn1, max=None ) )
        print( fmt( kn2, max=None ) )


        if self.doplot :
            plt.plot( xx, m1.result( xx, p1 ) )
            plt.plot( xx, m2.result( xx, p2 ) )
            plt.show()


    def constrainPos( self, logL, problem, allpars, logLlow ):
        xx = numpy.arange( 101, dtype=float )
        yy = problem.model.result( xx, allpars )
        if numpy.any( yy < 0 ):
            return logLlow - 1
        else :
            return logL

    def testNS( self ) :
        ky = numpy.arange( 100, dtype=int )
        py = numpy.asarray( [0,0,0,0.3,0.5,1.1,3.9,13.4,20.6,23.1] )

        perc = py[ ky // 10 ]
        year = numpy.asarray( ky, dtype=float )

        print( "  Test SplinesDynamicModel in NestedSampler" )

        # We need initial knot settings
        knots =[0, 50, 100]

        mdl = SplinesDynamicModel( knots=knots, maxKnots=5 )
        mdl.setLimits( lowLimits=[-20.0,-20.0], highLimits=[+30.0,30.0] )

        ns = NestedSampler( year, mdl, perc, seed=1234 )
        ns.distribution.setLimits( [0.01,100] )
        ns.distribution.constrain = self.constrainPos
        ns.verbose = 2
        if not self.dofull :
            ns.ensemble = 10

        evid = ns.sample( plot=self.doplot )

        sl = ns.samples
        ay = numpy.linspace( 0, 100, 101, dtype=float )
        res = sl.average( ay )
        self.assertTrue( all( res >= 0 ) )


    def growshrink( self, m0, dnp=1, npf=0, rng=None, prnt=False ) :
        Tools.printclass( m0 )

        self.assertTrue( m0.npchain == npf + dnp )
        self.assertTrue( m0.npbase == npf + dnp )
        self.report( m0, prnt=prnt )

        npf += dnp
        m = m0.copy()
        if m.grow( rng=rng ) :
            self.report( m, prnt=prnt )
            m0 = m
            npf += dnp
        else :
            print( "grow   failed" )

        m = m0.copy()
        if m.grow( rng=rng ) :
            self.report( m, prnt=prnt )
            m0 = m
            npf += dnp
        else :
            print( "grow   failed" )

        m = m0.copy()
        if m.grow( rng=rng ) :
            self.report( m, prnt=prnt )
            m0 = m
            npf += dnp
        else :
            print( "grow   failed" )

        print( m.npchain, m.npbase, m.npmax, npf )
        self.assertTrue( m.npchain == npf )
        self.assertTrue( m.npbase == npf )

        m = m0.copy()
        if m.grow( rng=rng ) :
            self.report( m, prnt=prnt )
            m0 = m
            npf += dnp
        else :
            print( "grow   failed" )

        m = m0.copy()
        if m.shrink( rng=rng ) :
            self.report( m, prnt=prnt )
            m0 = m
            npf -= dnp
        else :
            print( "shrink failed" )

        self.assertTrue( m.npchain == npf )
        self.assertTrue( m.npbase == npf )

        m = m0.copy()
        if m.shrink( rng=rng ) :
            self.report( m, prnt=prnt )
            m0 = m
            npf -= dnp
        else :
            print( "shrink failed" )

        self.assertTrue( m.npchain == npf )
        self.assertTrue( m.npbase == npf )

        m = m0.copy()
        if m.grow( rng=rng ) :
            self.report( m, prnt=prnt )
            m0 = m
            npf += dnp
        else :
            print( "grow   failed" )

        m = m0.copy()
        if m.grow( rng=rng ) :
            self.report( m, prnt=prnt )
            m0 = m
            npf += dnp
        else :
            print( "grow   failed" )

        return m

    def report( self, m, prnt=False ) :
        if prnt :
            print( "Knots  ", fmt( m.knots, max=None ) )
            print( "Param  ", fmt( m.parameters, max=None ) )

    def test2Model( self ):
        print( "========================" )
        print( "  Test two models" )
        print( "========================" )

        self.assertRaises( ValueError, PolynomialDynamicModel, 0, minDegree=2, maxDegree=4 )

        m = GaussModel( )
        p = PolynomialDynamicModel( 1, minDegree=1, maxDegree=4 )

        print( m.isDynamic(), p.isDynamic() )

        print( p.growPrior )
        m += p

        self.assertTrue( m.isDynamic() )
        self.assertFalse( isinstance( m, Dynamic ) )
        self.assertTrue( isinstance( m._next, Dynamic ) )

        lolim = [-10, -1, 0, -10]
        hilim = [+10, +1, 10, 10]
        m.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( m )
        Tools.printclass( p )

        par = numpy.asarray( [1,-0.2,0.3,-1.2,0.2] )
        x = numpy.linspace( -1.0, 1.0, 11 )
        y = numpy.linspace( 0.0, 2.3, 11 )

        problem = ClassicProblem( model=m, xdata=x, ydata=y )

        print( problem.myEngines() )
        print( problem.myStartEngine() )
        print( problem.myDistribution() )

        self.assertTrue( "galilean" in problem.myEngines() )
        self.assertTrue( "birth" in problem.myEngines() )
        self.assertTrue( "death" in problem.myEngines() )

        errdis = GaussErrorDistribution( limits=[0.01,100] )

        allpars = numpy.append( par, [1.0] )
        fitIndex = [0,1,2,3,4,-1]
        sl = WalkerList( problem, 10, allpars, fitIndex )

        print( m.npchain, m.npbase, m._next.npbase, len( allpars ) )
        print( "====================" )

        seng = StartEngine( sl, errdis )
        for k, s in enumerate( sl ) :
            print( "Walker  ", s.id, s.problem.npars, s.problem.model.npars, s.problem.model.npchain )
            seng.execute( k, 0.0 )

        logl = sl.getLogLikelihoodEvolution()
        lowl = numpy.min( logl )
        print( "logl  ", fmt( logl, max=None ) )
        print( "lowl  ", fmt( lowl ) )

        beng = BirthEngine( sl, errdis )
        deng = DeathEngine( sl, errdis )
        print( "start  np ", m.npchain )
        for k, s in enumerate( sl ) :
            Tools.printclass( s )
            suc = beng.execute( k, lowl )
            print( "Birth  ", suc, s.problem.npars, s.problem.model.npars, s.problem.model.npchain )
            Tools.printclass( s )
            suc = deng.execute( k, lowl )
            print( "Death  ", suc, s.problem.npars, s.problem.model.npars, s.problem.model.npchain )


    def testNestedSampler( self ) :
        print( "========================" )
        print( "  Test Nested Sampler" )
        print( "========================" )

        numpy.random.seed( 1315 )
        N = 21
        x = numpy.linspace( -2.0, 3.0, N, dtype=float )
        y = 1.2 + 0.5 * x + 0.33 * x * x + 0.27 * x * x * x
        noise = numpy.random.randn( N )
        y += 0.02 * noise

        pm = PolynomialDynamicModel( 0 )
        lolim = [-10]
        hilim = [+10]
        pm.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( pm )

        engines = None
        ns = NestedSampler( x, pm, y, seed=2031967, engines=engines )

        ns.distribution.setLimits( [0.01, 100] )

        if not self.dofull :
            ns.ensemble = 10

        logE = ns.sample( plot=self.doplot )


    def testNestedSampler2( self ) :
        print( "========================" )
        print( "  Test Nested Sampler 2" )
        print( "========================" )

        numpy.random.seed( 1315 )
        N = 40
        x = numpy.arange( N, dtype=float ) / 10
        y = [0, 1, 3, 4, 6, 7, 5, 8, 9, 2, -2,-9,-8,-5,-7,-6,-4,-3,-1, 0] * 2
        y - numpy.asarray( y, dtype=float )
#        y = numpy.linspace( 0.0, 4.0, N, dtype=float ) % 2
        noise = numpy.random.randn( N )
        y += 0.3 * noise

        pm = HarmonicDynamicModel( 1, period=2.0 )
        pm.growPrior.scale = 6

        lolim = [-10]
        hilim = [+10]
        pm.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( pm )

        ns = NestedSampler( x, pm, y, seed=2031967 )
#        ns.verbose = 4

        ns.distribution.setLimits( [0.01, 100] )
#        ns.distribution.scale = 0.02

        if not self.dofull :
            ns.ensemble = 10

        logE = ns.sample( plot=self.doplot )
        print( "scale  ", ns.scale )

    def testNestedSampler3( self ) :
        print( "========================" )
        print( "  Test Nested Sampler 3" )
        print( "========================" )

        numpy.random.seed( 1315 )
        N = 200
        x = numpy.arange( N, dtype=float ) / 40
        y = 0.1 * numpy.random.randn( N )
        gm = GaussModel()
        for k in range( 1, 6 ) :
            par = [5.0/k, k-0.5, 0.1]
            y += gm.result( x, par )

        pm = RepeatingModel( 1, gm )
        pm.growPrior.scale = 5
        lolim = [0.0,0.0,0.0]
        hilim = [+10,5.0,1.0]
        pm.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( pm )

        ns = NestedSampler( x, pm, y, seed=2031967 )
        ns.distribution.setLimits( [0.01, 100] )
#        ns.distribution.scale = 0.02

        if not self.dofull :
            ns.ensemble = 10

        logE = ns.sample( plot=self.doplot )
        sl = ns.samples
        print( "maxlik   ", fmt( sl.maxLikelihoodParameters, max=None ) )
        print( "median   ", fmt( sl.medianParameters, max=None ) )
        print( "modus    ", fmt( sl.modusParameters, max=None ) )
        print( "scale    ", fmt( ns.scale ) )


    def suite( cls ):
        return unittest.TestCase.suite( TestDynamicModel.__class__ )


if __name__ == '__main__':
    unittest.main()




import numpy as numpy
from . import Tools

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2020 Do Kester

class OrthonormalBasis( object ):
    """
    Helper class to construct a orthonormal basis from (random) vectors

    Attributes
    ----------
    basis : 2darray
        array of orthonormal vectors

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************

    def __init__( self ):
        """
        Constructor.
        """
        pass

    def normalise( self, vec, reset=False ) :
        """
        Construct from vec a unit vector orthogonal to self.basis

        from http://www.ecs.umass.edu/ece/ece313/Online_help/gram.pdf

        Parameters
        ----------
        vec : array_like
            vector to be orthonomalised to self.basis
        reset : bool
            start a new basis.
        """
        nb = 0 if reset or not hasattr( self, "basis" ) else self.basis.shape[0]
        if nb >= len( vec ) :
            nb = 0

        uv = 0.0
        for k in range( nb ) :
            u = self.basis[k,:]
            uv += numpy.inner( u, vec ) * u
        uvec = vec - uv
        uvec /= numpy.linalg.norm( uvec )

        if nb == 0 :
            self.basis = uvec.reshape( (1,-1) )
        else :
            self.basis = numpy.append( self.basis, [uvec], 0 )

#        print( nb, self.basis.shape )
        return uvec


    def __str__( self ) :
        return str( "OrthonormalBasis" )




import numpy as numpy
from . import Tools
from .Formatter import formatter as fmt
from .Tools import setAttribute as setatt
from .Prior import Prior
from .ExponentialPrior import ExponentialPrior
from .UniformPrior import UniformPrior

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2021 Do Kester


class Dynamic( object ):
    """
    Class adjoint to Model which implements some dynamic behaviour.


    Attributes
    ----------
    ncomp : int
        the number of components in the dynamic model
    deltaNpar : int
        the number of parameters in each component
    minComp : int
        minimum number of repetitions
    maxComp : None or int
        maximum number of repetitions
    growPrior : None or Prior
        governing the birth and death.
        ExponentialPrior (scale=2) if  maxOrder is None else UniformPrior

    """
    def __init__( self, dynamic=True ) :
        """
        Constructor for Dynamic

        Parameters
        ----------
        dynamic: bool
            True if the Model is to be considered dynamic.
        """

        setatt( self, "dynamic", dynamic )


    def isDynamic( self ) :
        return self.dynamic

    def setGrowPrior( self, growPrior=None, min=1, max=None, name="Comp" ) :
        """
        Set the growth prior.

        Parameters
        ----------
        growPrior : None or Prior
            governing the birth and death.
            ExponentialPrior (scale=2) if  maxOrder is None else UniformPrior
        min : int
            lower limit on growthprior
        max : None or int
            upper limit on growthprior
        name : str
            name of the component
        """
        setatt( self, "min%s"%name, min, type=int )
        setatt( self, "max%s"%name, max, type=int, isnone=True )
        if growPrior is None :
            if max is None :
                setatt( self, "growPrior", ExponentialPrior( scale=2 ) )
            else :
                lim = [min, max+1]              # limits on components
                setatt( self, "growPrior", UniformPrior( limits=lim ) )
        else :
            setatt( self, "growPrior", growPrior, type=Prior )


    def setDynamicAttribute( self, name, value ) :
        """
        Set attribute, if it belongs to a Dynamic Models.

        Parameters
        ----------
        name : str
            name of the attribute
        value : anything
            value of the attribute

        Return
        ------
        bool : True if name was a Dynamic name
               False if not

        """
        if name == "ncomp" :
            setatt( self, name, value, type=int )
            return True
        if name == "deltaNpar" :
            setatt( self, name, value, type=int )
            return True
        if name == "maxComp" :
            setatt( self, name, value, type=int, isnone=True )
            return True
        if name == "minComp" :
            setatt( self, name, value, type=int )
            return True
        if name == "growPrior" :
            setatt( self, name, value, type=Prior )
            return True

        return False


    def grow( self, offset=0, rng=None, **kwargs ):
        """
        Increase the degree by one upto maxComp ( if present ).

        Parameters
        ----------
        offset : int
            index where the params of the Dynamic model start
        rng : random number generator
            to generate a new parameter.

        Return
        ------
        bool :  succes

        """
        if self.maxComp is not None and self.ncomp >= self.maxComp:
            return False

        location = self.npbase

        dnp = self.deltaNpar
        self.alterParameterSize( dnp, offset, location=location )

        head = self._head
        mdlpar = head.parameters
        k1 = location + offset
        value = numpy.zeros( dnp, dtype=float )
        if rng is not None :
            for k in range( dnp ) :
                value[k] = head.unit2Domain( rng.rand(), k + k1 )

        mdlpar = self.alterParameters( mdlpar, location, dnp, offset, value=value )

        setatt( self._head, "parameters", mdlpar )

        self.alterParameterNames( dnp )

        self.changeNComp( 1 )

        return True

    def shrink( self, offset=0, rng=None, **kwargs ):
        """
        Decrease the degree by one downto minComp ( default 1 ).

        Parameters
        ----------
        offset : int
            index where the params of the Dynamic model start
        rng : random number generator
            Not used in this implementation

        Return
        ------
        bool : succes

        """
        if self.ncomp <= self.minComp :
            return False

        location = self.npbase
        dnp = - self.deltaNpar
        self.alterParameterSize( dnp, offset, location=location )

        head = self._head
        mdlpar = head.parameters
        mdlpar = self.alterParameters( mdlpar, location, dnp, offset )
        setatt( self._head, "parameters", mdlpar )

        self.alterParameterNames( dnp )

        self.changeNComp( -1 )

        return True

    def alterParameterNames( self, dnp ) :
        """
        Renumber the parameter names.

        Parameters
        ----------
        dnp : int
            change in the number of parameters
        """
        if dnp > 0 :
            parNames = ["parameter_%d"%k for k in range( self.npbase )]
        else :
            parNames = self.parNames[:self.npbase]

        setatt( self, "parNames", parNames )


    def alterParameterSize( self, dnp, offset, location=None, value=0 ) :
        """
        Change the number of parameters and self.parameters.

        Parameters
        ----------
        dnp : int
            change in the number of parameters in the DynamicModel
        offset : int
            starting index of the DynamicModel
        location : int
            index in param[offset:] at which to insert/delete the new parameters
        """
        if dnp == 0 :
            return

        if location is None :
            location = self.npbase

#        print( "DYN1  ", dnp, offset, self.npmax, self.npbase, self.npchain, mdlpar )

        setatt( self, "location", location )
        setatt( self, "npmax", self.npmax + dnp )
        setatt( self, "npbase", self.npbase + dnp )
        setatt( self._head, "_npchain", self._head._npchain + dnp )

#        print( "DYN2  ", dnp, offset, self.npmax, self.npbase, self.npchain, mdlpar )


    def alterParameters( self, param, location, dnp, offset, value=None ) :
        """
        change the parameters to comply with the changed model.

        param:      [p0 p1 p2 p3 p4 p5 p6 p7 p8 p9]   # previous set
        offset:     2           # parameters of models in preceeding chain
        location:   1           # location where to add/delete parameter
        value:      [v0 ...]    # values to be given to added parameters

        dnp:        +1
        ==> newpar: [p0 p1 p2 v0 p3 p4 p5 p6 p7 p8 p9]

        dnp:        +2
        ==> newpar: [p0 p1 p2 v0 v1 p3 p4 p5 p6 p7 p8 p9]

        dnp:        -1
        ==> newpar: [p0 p1 p3 p4 p5 p6 p7 p8 p9]

        dnp:        -2
        ==> ERROR: not enough space in param before location

        Parameters
        ----------
        param : array_like
            parameters of the parent model (chain)
        location : int
            index in param[offset:] at which to insert/delete the new parameters
        dnp : int
            number of parameters to insert (dnp>0) or delete (dnp<0)
        offset : int
            start index of the parameters of the dynamic model in param
        value : float or array_like
            to be given to the inserted parameters (only when dnp>0)

        """

#        print( "DY IN  ", fmt( param, max=None ), location, dnp, offset )

        np = len( param )
        newpar = numpy.zeros( np + dnp, dtype=float )

        if dnp > 0 :
            k1 = offset + location
            k2 = k1 + dnp
            newpar[:k1] = param[:k1]
            newpar[k1:k2] = value
            newpar[k2:] = param[k1:]
        else :
            if location + dnp < 0 :
                raise IndexError( "Cannot delete %d from parameters at %d"%(-dnp,location) )
            k2 = offset + location
            k1 = k2 + dnp
            newpar[:k1] = param[:k1]
            newpar[k1:] = param[k2:]

#        print( "DY OUT ", fmt( newpar, max=None ), k1, k2 )

        return newpar


    def alterFitindex( self, findex, location, dnp, offset ) :
        """
        change the fit index to comply with the changed model.

        Parameters
        ----------
        findex : array_like
            fit index of the parent model (chain)
        location : int
            index in param[offset:] at which to insert/delete the new parameters
        dnp : int
            number of parameters to insert (dnp>0) or delete (dnp<0)
        offset : int
            start index of the parameters of the dynamic model in param
        """
        lfin = len( findex )
        newfi = numpy.zeros( lfin + dnp, dtype=int )

        k = 0
        kol = offset + location

        kh = kol if dnp > 0 else kol + dnp
#        while findex[k] >=0 and findex[k] < kh :    ## copy first parts
        while k < lfin and findex[k] >=0 and findex[k] < kh :    ## copy first parts
            newfi[k] = findex[k]
            k += 1
        kl = k if dnp > 0 else k - dnp              ## where we are if findex

        for i in range( dnp ) :                     ## insert extra indices if dnp>0
            newfi[k] = kol + i
            k += 1

        for j in range( kl, lfin ) :                 ## copy the tail
            if findex[j] >= 0 :
                newfi[k] = findex[j] + dnp           ## in/de-crease indices
            else :
                newfi[k] = findex[j]                 ## negitive indices for hyperpars
            k += 1

        return newfi


    def shuffle( self, param, offset, np, rng ) :
        """
        Shuffle the parameters of the components (if they are equivalent)
        Default implementation: does nothing.

        Parameters
        ----------
        param : array-like
            list of all parameters
        offset : int
            index where the dynamic model starts
        np : int
            length of the parameters of the dynamic model
        rng : RNG
            random number generator
        """
        return param







import numpy as numpy
from .LinearModel import LinearModel
from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2004 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester

class PowerModel( LinearModel ):
    """
    General power model of arbitrary degree.

        f( x:p ) = p * x^a

    a is an float ( positive or negative ).

    To get the overall units of f(x:p) right, please note that the x-term
    needs to be divided by a factor 1.0 in the same units as the x.
    Otherwise possibly fractional dimensions are created.
    The factor is omitted as it does not contribute in the calculations.

    Examples
    --------
    >>> pwr = PowerModel( -1 )
    >>> print pwr.getNumberOfParameters( )       # 1
    1

    Attributes
    ----------
    exponent : float
        exponent (a) of the power function.

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    """
    def __init__( self, exponent=0, copy=None, **kwargs ):
        """
        Power of a certain degree.

        The number of parameters is 1

        Parameters
        ----------
        exponent : int
            power to which the xdata is to be raised.
        copy : PowerModel
            to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        param = [1.0]
        names = ["coefficient"]
        super( PowerModel, self ).__init__( 1, copy=copy, params=param,
                    names=names, **kwargs )

        if copy is None :
            self.exponent = float( exponent )
        else :
            self.exponent = copy.exponent

    def copy( self ):
        """ Copy method.  """
        return PowerModel( copy=self )

    def __setattr__( self, name, value ) :
        if name == "exponent" :
            setatt( self, name, value, type=float )
        else :
            super( PowerModel, self ).__setattr__( name, value )


    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the xdata value.
        <br>
        The partials are x ( xdata ) to degree-th power.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters. (not used for linear models)
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        if parlist is None or parlist[0] == 0 :
            inp = numpy.asarray( [xdata] ).transpose()
            return numpy.power( inp, self.exponent )
        else :
            return numpy.asarray( [] )

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        return params[0] * self.exponent * numpy.power( xdata, self.exponent - 1 )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "Power: f( x:p ) = p_0 * x^%.1f"%self.exponent )

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.
        Not strictly OK. See Class documentation.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        return self.yUnit





import numpy as numpy
import math

from .Formatter import formatter as fmt
from .Formatter import fma
from .ErrorDistribution import ErrorDistribution

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022 Do Kester


class BernoulliErrorDistribution( ErrorDistribution ):
    """
    To calculate a Bernoulli likelihood for categorical True/False data.

    For one residual, x, it holds

        f( x ) = x          if d is True
                 1 - x      if d is False

    where x needs to be between [0,1]; use the logistic function f(x) = 1/(1+exp(-x)
    if necessary. And d is true if the residual belongs to the intended category.

    The function is mostly used to calculate the likelihood L, or easier
    to use log likelihood, logL.

        logL = log( x ) if d else log( 1 - x )

    Author       Do Kester.

    """
    SQRT2 = math.sqrt( 2 )
    LGSQ2 = math.log( SQRT2 )
    LOG2 = math.log( 2.0 )


    #  *********CONSTRUCTORS***************************************************
    def __init__( self, copy=None ) :
        """
        Constructor of Bernoulli Distribution.

        Parameters
        ----------
        copy : BernoulliErrorDistribution
            distribution to be copied.
        """
        super( BernoulliErrorDistribution, self ).__init__( copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return BernoulliErrorDistribution( copy=self )

    #  *********DATA & WEIGHT***************************************************
    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return False

    def getScale( self, problem, allpars=None ) :
        """
        Return the noise scale

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        res = self.getResiduals( problem, allpars=allpars )
        return numpy.max( numpy.abs( res ) )

    def toSigma( self, scale ) :
        """
        Return sigma, the squareroot of the variance.
        Parameter
        --------
        scale : float
            the scale of this Bernoulli distribution.
        """
        return scale * 2 / math.sqrt( 12 )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ) :
        """
        Return the log( likelihood ) for a Bernoulli distribution.

        Alternate calculation.

        Outside the range the likelihood is zero, so the logL should be -inf.
        However for computational reasons the maximum negative value is returned.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem

        """
        self.ncalls += 1

        return numpy.sum( self.logLdata( problem, allpars ) )

    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for each residual

        logL = sum( logLdata )

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        mockdata : array_like
            as calculated by the model

        """
        if mockdata is None :
            mockdata = problem.result( allpars )

        if problem.model.lastndout == 1 :
            return numpy.log( numpy.where( problem.ydata == 0, mockdata, 1 - mockdata ) )

        lld = numpy.zeros( len( problem.ydata ), dtype=float )

        for k in range( problem.model.lastndout ) :
            lld += numpy.where( problem.ydata == k, mockdata[:,k], 0 )

        return numpy.log( lld )

    def partialLogL_alt( self, problem, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        dL = numpy.zeros( len( fitIndex ), dtype=float )

        mock = problem.result( allpars )
        dM = problem.partial( allpars )
        dLdM = numpy.zeros_like( mock )

        if problem.model.lastndout == 1 :
            dLdM = 1.0 / numpy.where( problem.ydata == 0, mock, 1 - mock )
            dL = numpy.sum( numpy.multiply( dLdM, dM.T ), axis=1 )
            return dL

        for k in range( problem.model.lastndout ) :
            q = numpy.where( problem.ydata == k )
            dLdM[q,k] = 1.0 / mock[q,k]

        dL = numpy.tensordot( dLdM, numpy.asarray( dM ), axes=([0,1],[1,0]) )

        return dL

    def nextPartialData( self, problem, allpars, fitIndex, mockdata=None ) :
        """
        Return the partial derivative of elements of the log( likelihood )
        to the parameters.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted
        mockdata : array_like
            as calculated by the model

        """
        if mockdata is None :
            mock = problem.result( allpars )
        else :
            mock = mockdata

        dM = problem.partial( allpars )
        dLdM = numpy.zeros_like( mock )

        if problem.model.lastndout == 1 :
            dLdM = 1.0 / numpy.where( problem.ydata == 0, mock, mock - 1 )
            dL = numpy.multiply( dLdM, dM.T )

        else :
            for k in range( problem.model.lastndout ) :
                q = numpy.where( problem.ydata == k )
                dLdM[q,k] = 1.0 / mock[q,k]

            dL = numpy.tensordot( dLdM, numpy.asarray( dM ), axes=([0,1],[1,0]) )

#        r0 = numpy.log( numpy.where( problem.ydata == 0, mock - 0.001, 1 - mock + 0.001 ) )
#        r1 = numpy.log( numpy.where( problem.ydata == 0, mock + 0.001, 1 - mock - 0.001 ) )


        for k in fitIndex :
            yield dL[k]

        return

    def __str__( self ) :
        return "Bernoulli error distribution"




import numpy as numpy
from astropy import units
import re
import warnings
from . import Tools

from .Problem import Problem
from .Dynamic import Dynamic
from .Modifiable import Modifiable

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2023 Do Kester

class ClassicProblem( Problem ):
    """
    A ClassicProblem is an optimization of parameters which involves
    the fitting of data to a Model at a fixed set of x values.

    Problems can be solved by NestedSampler, with appropriate Engines and
    ErrorDistributions.

    The result of the function for certain x and p is given by
    problem.result( x, p )
    The parameters, p, are to be optimized while the x provide additional
    information.

    Attributes from Problem
    -----------------------
    model, xdata, ydata, weights, accuracy, varyy

    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, model=None, xdata=None, ydata=None, weights=None, 
                    accuracy=None, copy=None ):
        """
        Constructor for classic problems.

        Parameters
        ----------
        model : Model
            the model to be solved
        xdata : array_like or None
            independent variable
        ydata : array_like or None
            dependent variable
        weights : array_like or None
            weights associated with ydata
        accuracy : float or array_like
            accuracy scale for the datapoints
            all the same or one for each data point
        copy : Problem
            to be copied

        """
        super( ).__init__( model=model, xdata=xdata, ydata=ydata, weights=weights,
                           accuracy=accuracy, copy=copy )


    def copy( self ):
        """
        Copy.

        The copy points to the same instance of model.
        """
        return ClassicProblem( copy=self )

    #  *****RESULT**************************************************************
    def result( self, param ):
        """
        Returns the result calculated at the xdatas.

        Parameters
        ----------
        param : array_like
            values for the parameters.

        """
        return self.model.result( self.xdata, param )


    def partial( self, param ) :
        """
        Return the partials of the internal model.

        Parameters
        ----------
        param : array_like
            list of model parameters
        """
        return self.model.partial( self.xdata, param )

    def derivative( self, param ) :
        """
        Return the derivative of the internal model.

        Parameters
        ----------
        param : array_like
            list of model parameters
        """
        return self.model.derivative( self.xdata, param )

    def myEngines( self ) :
        """
        Return a default list of preferred engines
        """
        engs = ["galilean", "chord"]

        if self.model.isDynamic() :
            engs += ["birth", "death"]

        if self.model.isModifiable() :
            engs += ["struct"]

        return engs

    def myStartEngine( self ) :
        """
        Return a default preferred start engines: "start"
        """
        return "start"

    def myDistribution( self ) :
        """
        Return a default preferred ErrorDistribution: "gauss"
        """
        return "gauss"

    #  *****TOSTRING***********************************************************
    def baseName( self ):
        """ Returns a string representation of the model.  """
        return "ClassicProblem of " + self.model.__str__()






import time
import numpy as numpy
from astropy import units
import math
from . import Tools
import matplotlib.pyplot as pyplot

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester

class IterationPlotter( object ):
    """
    The IterationPlotter plots intermediate results from a iterative fitter.

    Author:      Do Kester

    """
    def plotData( self, x, y, title ):
        """
        Plot the data.

        Parameters
        ----------
        x : array_like
            x-axis values of the data
        y : array_like
            y-axis values of the data
        title : string
            the title of the plot
        """

        self.p = pyplot.gca()
        self.p.plot( x, y, 'k.' )
        pyplot.show( block=False )
        time.sleep( 1 )

    def plotResult( self, x, r, iter ):
        """
        Plot the ( intermediate ) result.

        Parameters
        ----------
        x : array_like
            x-axis values of the data
        r : array_like
            model result
        iter : int
            iteration number
        """
        self.p.plot( x, r, 'r-' )
        pyplot.show( block=False )
        time.sleep( 1 )

    def plotProgress( self, percent ):
        """
        Plot (estimated) progress upto now.

        """
        pass





import numpy as numpy
from . import Tools
from .Formatter import formatter as fmt
from .Formatter import fma

from .Engine import Engine
from .Engine import DummyPlotter

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester

class StepEngine( Engine ):
    """
    Move a walker in a random direction.

    The StepEngine tries to move the parameters at most 4 times in
    a random direction.

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213, verbose=0 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : WalkerList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : StepEngine
            to be copied
        seed : int
            for rng

        """
        super( ).__init__( walkers, errdis, copy=copy, seed=seed, verbose=verbose )
        self.plotter = DummyPlotter()


    def copy( self ):
        """ Return copy of this.  """
        return StepEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "StepEngine" )


    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        kw : int
            index of walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place or append
        iteration : int
            iteration number

        Returns
        -------
        int : the number of successfull steps

        """
        self.reportCall()

        walker = self.walkers[kw].copy()
        problem = walker.problem
        fitIndex = walker.fitIndex
        np = len( fitIndex )

        urange = self.unitRange[fitIndex]
        dur = urange / len( self.walkers )      ## why this ???
        urange += 2 * dur

        param = walker.allpars
        self.plotter.start( param=param )
        usav = self.domain2Unit( problem, param[fitIndex], kpar=fitIndex )

        if self.verbose > 4 :
            print( "alpar ", fma( param ), fmt( walker.logL ), fmt( lowLhood) )
            fip = param[fitIndex]
            print( "uap   ", fma( self.domain2Unit( problem, fip, fitIndex )))
            print( "fitin ", fma( fitIndex ), self.maxtrials )
            print( "unitr ", fma( self.unitRange ) )


        sz = 1.0
        step = ( self.rng.rand( np ) - 0.5 ) * urange
        kk = 0

        while kk < 4 :
            ptry = param.copy()

            while True :
                utry = usav + step
                q0 = numpy.where( utry < 0 )[0]
                nq0 = len( q0 )
                q1 = numpy.where( utry > 1 )[0]
                nq1 = len( q1 )

                if nq0 > 0 :
                    step[q0] = ( self.rng.rand( nq0 ) - 0.5 ) * urange[q0]
                elif nq1 > 0 :
                    step[q1] = ( self.rng.rand( nq1 ) - 0.5 ) * urange[q1]
                else :
                    break

            ptry[fitIndex] = self.unit2Domain( problem, utry, kpar=fitIndex  )

            Ltry = self.errdis.logLikelihood( problem, ptry )

            if self.verbose > 4 :
                print( "Step  ", fma(ptry), fmt(Ltry), kk, fmt( sz ) )

            if Ltry >= lowLhood:
                self.plotter.move( param, ptry, col=0, sym=0 )
                self.reportSuccess( )

                update = len( self.walkers ) if append else kw
                self.setWalker( update, problem, ptry, Ltry, fitIndex=fitIndex )
                param = ptry
                usav = utry
                sz = 1.0
                step = ( self.rng.rand( np ) - 0.5 ) * urange
                kk += 1
#                break
            elif kk <= self.maxtrials :
                self.plotter.move( param, ptry, col=5, sym=4 )
                sz *= 0.7
                step *= sz
                self.reportReject( )
            else :
                self.reportFailed()
                break

        self.plotter.stop( param=param, name="StepEngine" )

        return kk                        # nr of succesfull steps





# run with : python3 -m unittest TestAmoeba

import numpy as numpy
import os
import unittest
from astropy import units
import math
import matplotlib.pyplot as plt
from numpy.testing import assert_array_almost_equal as assertAAE

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2002 Do Kester

class TestAmoeba( unittest.TestCase ):
    """
    Test harness for AnnealingAmoeba class.

    Author       Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot =  ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def func1( self, xx ) :
        yy = 3.14
        p = 1.2
        for x in xx :
            x += p
            yy += x * x
            p *= p
        return yy

    def func2( self, xx ) :
        yy = 3.14
        p = 1.2
        a = 1.0 * math.pi
        for x in xx :
            x += p
            yy += x * ( x + 1 ) * ( 2 - numpy.cos( a * x ) )
            p *= p
        return yy

    #  **************************************************************
    def test1( self ):
        """     # 	test iterative slope fit  """
        print( "Testing AnnealingAmoeba; test 1" )

        xx = numpy.asarray( [3, -2], dtype=float )
        amb = AnnealingAmoeba( self.func1, xx, verbose=1 )

        Tools.printclass( amb )
        amb.minimize()

        assertAAE( amb.xopt, [-1.2,-1.44], 2 )
        assertAAE( amb.fopt, self.func1( [-1.2,-1.44] ), 3 )

    #  **************************************************************
    def test2( self ):
        """     # 	test iterative slope fit  """
        print( "Testing AnnealingAmoeba with limits; test2" )

        xx = numpy.asarray( [3, -2], dtype=float )
        amb = AnnealingAmoeba( self.func1, xx, limits=[[-1,0],[3,3]], reltol=1e-10,
            maxiter=100, verbose=1 )
        Tools.printclass( amb )
        print( "Catch runtime error" )
        self.assertRaises( ConvergenceError, amb.minimize )

        print( "Testing AnnealingAmoeba with limits; test2" )
        xx = numpy.asarray( [3, -2], dtype=float )
        amb = AnnealingAmoeba( self.func1, xx, limits=[[-1,0],[3,3]], reltol=1e-10,
                verbose=5 )

        amb.minimize()

        assertAAE( amb.xopt, [-1.0,0.0] )
        assertAAE( amb.fopt, self.func1( [-1.0,0.0] ) )

    #  **************************************************************
    def test3( self ):
        """     # 	test iterative slope fit  """
        print( "Testing AnnealingAmoeba with limits; test 3" )

        if self.doplot :
            x = numpy.linspace( -10, 10, 201, dtype=float )
            y = numpy.linspace( -10, 10, 201, dtype=float )
            self.plot( self.func2, x=x, y=y, contour=[5,10,15,20,30,40,60,80,100, 200] )

        xx = numpy.asarray( [3, -2], dtype=float )
        amb = AnnealingAmoeba( self.func2, xx, reltol=1e-10, verbose=1 )
        xopt = amb.minimize()

        print( amb.fopt, xopt )
        if self.doplot :
            plt.plot( [xopt[1]], [xopt[0]], 'r+' )


        amb = AnnealingAmoeba( self.func2, xx, temp=10, seed=123456, verbose=1 )
        amb.minimize()

        xopt = amb.xopt
        print( amb.fopt, xopt )
        if self.doplot :
            plt.plot( [xopt[1]], [xopt[0]], 'g+' )
            plt.show()

        assertAAE( amb.xopt, [-1.836, -2.0767], 3 )
        assertAAE( amb.fopt, 2.0221211, 5 )
# 2.02212110215 [-1.83671038 -2.07670877]



    def plot( self, func, x=None, y=None, contour=None ) :
        if x is None :
            x = numpy.linspace( -10, 10, 21 )
        if y is None :
            y = numpy.linspace( -10, 10, 41 )
        nx = len( x )
        ny = len( y )

        map = numpy.ndarray( (nx,ny), dtype=float )
        k0 = 0
        for yy in y :
            k1 = 0
            for xx in x :
                map[k1,k0] = func( [xx,yy] )
                k1 += 1
            k0 += 1

        if contour is None :
            contour = 10
        plt.contour( x, y, map, contour )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestAmoeba.__class__ )


if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
import math
import re
import sys

from .ErrorDistribution import ErrorDistribution
from .HyperParameter import HyperParameter
from .UniformPrior import UniformPrior
from . import Tools
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022 Do Kester


class MixedErrorDistribution( ErrorDistribution ):
    """
    To calculate a mixture of two likelihoods.

    For one residual, x, it holds

        L( x ) = f * L1( x ) + ( 1 - f ) * L2( x )

    where f is the contributing fraction while L, L1 and L2 are likelihoods
    f is a hyperparameter between [0..1]

    The likelihood over N datapoints is

        L = &Pi;{ L( x ) } = &Pi;( f * L1( x ) + ( 1 - f ) * L2( x ) )

    And the log of L is

        logL = &sum; logL( x ) = &sum;( log( f * L1(x) + ( 1 - f ) * L2(x) ) )

    Note
    ----
    The mixture, i.e. the weighted sum of 2 distributions for each residual, is
    the raison-d'etre for the methods logLdata and nextPartialData, so individual
    contributions can be weighted, added, log-ged and summed.

    Author       Do Kester.

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, errdis1, errdis2, fraction=0.5, limits=None, copy=None ):
        """
        Constructor.

        Make a new error distribution as a fraction of errdis1 plus the rest of errdis2.

        Parameters
        ----------
        errdis1 : ErrorDistribution
            First error distribution
        errdis2 : ErrorDistribution
            Second error distribution (might be of the same class as errdis1)
            It *must* have the same xdata, data, weights as errdis1.
        fraction : float
            contributing fraction
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed fraction
            low     low limit on fraction ( >0)
            high    high limit on fraction ( <1)
            when limits are set, the scale is *not* fixed.

        copy : MixedErrorDistribution
            distribution to be copied.

        """
        super( MixedErrorDistribution, self ).__init__( copy=copy )

        self.errdis1 = errdis1
        self.errdis2 = errdis2
        if copy is None :
            self.hyperpar = errdis1.hyperpar + errdis2.hyperpar
            fprior = UniformPrior( )
            fraction = HyperParameter( hypar=fraction, prior=fprior, limits=limits )
            self.hyperpar += [fraction]

    def copy( self ):
        """ Return copy of this.  """
        return MixedErrorDistribution( self.errdis1, self.errdis2, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        key1 = {"errdis1": ErrorDistribution, "errdis2" : ErrorDistribution }

        if Tools.setSingleAttributes( self, name, value, key1 ) :
            pass
        else :
            super( MixedErrorDistribution, self ).__setattr__( name, value )


    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return self.errdis1.acceptWeight() and self.errdis2.acceptWeight()


    #  *********LIKELIHOODS***************************************************
    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for a Mixedian distribution.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        mockdata : array_like
            as calculated for the problem

        """
        if mockdata is None :
            mockdata = problem.result( allpars[:problem.npars] )

        f = allpars[-1]
        n1 = self.errdis1.nphypar
        n2 = self.errdis2.nphypar
        p1 = allpars[:-n2-1]
        p2 = allpars[:-n1-1].copy()
        p2[-n2:] = allpars[-n2-1:-1]

        if f <= 0 :
            return self.errdis2.logLdata( problem, p2, mockdata=mockdata )
        if f >= 1 :
            return self.errdis1.logLdata( problem, p1, mockdata=mockdata )

        return numpy.logaddexp(
                self.errdis1.logLdata( problem, p1, mockdata=mockdata ) + math.log( f ),
                self.errdis2.logLdata( problem, p2, mockdata=mockdata ) + math.log( 1 - f ) )

    def nextPartialData( self, problem, allpars, fitIndex, mockdata=None ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters in fitIndex.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted
        mockdata : array_like
            as calculated for the problem

        """
        if mockdata is None :
            mockdata = problem.result( allpars[:problem.npars] )

        # make allpars lists (p1,p2)for errdis1&2; shift hyperpars in p2
        f = allpars[-1]
        emf = 1 - f
        n1 = self.errdis1.nphypar
        n2 = self.errdis2.nphypar
        p1 = allpars[:-n2-1]
        p2 = allpars[:-n1-1].copy()
        p2[-n2:] = allpars[-n2-1:-1]


        # make fitIndex (f1,f2) for errdis1&2
        fitIndex = numpy.asarray( fitIndex )
        q = numpy.where( fitIndex >= 0 )
        f1 = fitIndex[q]
        f2 = f1.copy()
        if -3 in fitIndex :
            f1 = numpy.append( f1, [-1] )
        if -2 in fitIndex :
            f2 = numpy.append( f2, [-1] )

        # make fitindices (f1,f2) for errdis1&2
        lhd1 = numpy.exp( self.errdis1.logLdata( problem, p1, mockdata=mockdata ) )
        lhd2 = numpy.exp( self.errdis2.logLdata( problem, p2, mockdata=mockdata ) )
        pg1 = self.errdis1.nextPartialData( problem, p1, f1, mockdata=mockdata )
        pg2 = self.errdis2.nextPartialData( problem, p2, f2, mockdata=mockdata )

        ff1 = f * lhd1 + emf * lhd2

        # Where the mixed likelihood (=ff1) is indistinguishable from 0,
        # its partials also must be zero. In those locations we can replace
        # the inverse of ff1, needed in calculating the partials, by 0.
        fff = numpy.zeros_like( ff1 )
        q = numpy.where( ff1 > sys.float_info.min )
        fff[q] = 1.0 / ff1[q]

#        fff = numpy.where( ff1 < sys.float_info.min, 0.0, 1.0 / ff1 )

        for fi in fitIndex :
            if fi >= 0 :            ## partialLogL for model parameters
                npg1 = next( pg1 )
                npg2 = next( pg2 )

                yield ( fff * ( f * lhd1 * npg1 +
                              emf * lhd2 * npg2 ) )

            if fi == -1 :           ## partialLogL for fraction f
                yield ( ( lhd1 - lhd2 ) * fff )

            elif fi == -2 :         ## partialLogL for sigma of errdis2
                npg2 = next( pg2 )

                yield ( emf * lhd2 * fff * npg2 )

            elif fi == -3 :         ## partialLogL for sigma of errdis1
                npg1 = next( pg1 )

                yield ( f * lhd1 * fff * npg1 )


    def __str__( self ) :
        n1 = re.match( "^[a-zA-Z_]*", self.errdis1.__str__() ).group(0)
        n2 = re.match( "^[a-zA-Z_]*", self.errdis2.__str__() ).group(0)
        return "Mixed error distribution : %s + %s" % (n1, n2)



import numpy as numpy
from . import Tools

from .Walker import Walker

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2021 Do Kester

class Engine( object ):
    """
    Engine defines common properties of all Engines.

    An Engine moves around a walker in a random way such that its likelood
    remain above the low-likelihood-limit.

    Attributes
    ----------
    walkers : WalkerList
        list of walkers to be diffused
    errdis : ErrorDistribution
        error distribution to be used
    slow : int
        If slow > 0, run this engine every slow-th iteration.
    maxtrials : int
        maximum number of trials for various operations
    rng : numpy.random.RandomState
        random number generator
    verbose : int
        if verbose > 4 report about the engines. (mostly for debugging)

    report : list of int (read only)
        reports number of succes, accepted, rejected, failed calls. Plus the total.
    unitRange : array_like (read only)
        present max size of the parameter cloud (in unitspace: [0,1])
    unitMin : array_like (read only)
        present minimum values of the parameter cloud (in unitspace: [0,1])

    Author       Do Kester.

    """
    SUCCESS = 0             #  succesfull move
    REJECT = 1              #  rejected move
    FAILED = 2              #  failed to move
    BEST   = 3              #  better than all
    NCALLS = 4              #  number of calls

    #  *********CONSTRUCTORS***************************************************

    def __init__( self, walkers, errdis, slow=None, copy=None, seed=4213, verbose=0 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : list of Walker
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        slow : None or int > 0
            Run this engine every slow-th iteration. None for all.
        seed : int
            for random number generator
        verbose : int
            report about the engines when verbose > 4
        copy : Engine
            engine to be copied

        """
        self.walkers = walkers
        self.errdis = errdis
        self.report = [0]*5

        if copy is None :
            self.maxtrials = 5
            self.rng = numpy.random.RandomState( seed )
            self.unitRange = None
            self.unitMin = None
            self.verbose = verbose
            if slow is not None : self.slow = slow
            # self.setWalker = self.setWalkerAdd2List if usePhantoms else self.setWalkerInPlace
            # self.lastWalkerId = 0
        else :
            self.maxtrials = copy.maxtrials
            self.rng = copy.rng
            self.unitRange = copy.unitRange
            self.unitMin   = copy.unitMin
            self.verbose   = copy.verbose
            if hasattr( copy, "slow" ) : self.slow = copy.slow
            # self.setWalker = copy.setWalker
            # self.lastWalkerId = copy.lastWalkerId

    def copy( self ):
        """ Return a copy of this engine.  """
        return Engine( self.walkers, self.errdis, copy=self )

    #  *********SET & GET***************************************************
    def setWalker( self, kw, problem, allpars, logL, fitIndex=None ) :
        """
        Update the walker with problem, allpars, LogL and logW.

        Parameters
        ----------
        walker : Sample
            sample to be updated

        kw : int
            index in walkerlist, of the walker to be replaced
        problem : Problem
            the problem in the walker
        allpars : array_like
            list of all parameters
        logL : float
            log Likelihood
        fitIndex : array_like
            (new) fitIndex
        """
        id = 0 if kw >= len( self.walkers ) else self.walkers[kw].id
        walker = Walker( id, problem, allpars, fitIndex )
        walker.logL = logL
        self.walkers.setWalker( walker, kw )

#       DONT DO THIS ANY MORE        
#        self.checkBest( problem, allpars, logL, fitIndex=fitIndex )


    def checkBest( self, problem, allpars, logL, fitIndex=None ) :
        """
        Check if Ltry better than the best at self.walkers[-1].
        If so replace.

        Parameters
        ----------
        problem : Problem
            the problem in the walker
        logL : float
            likelihood
        allpars : array_like
            parameters of problem
        fitIndex : array_like
            (new) fitIndex
        """
        if logL > self.walkers[-1].logL :
            self.setWalker( -1, problem, allpars.copy(), logL, fitIndex )
            self.reportBest()

######## domain <> unit ###########################################

    def domain2Unit( self, problem, dval, kpar=None ) :
        """
        Return value in [0,1] for the selected parameter.

        Parameters
        ----------
        problem : Problem
            the problem involved
        dval : float
            domain value for the selected parameter
        kpar : None or array_like
            selected parameter index, where kp is index in [parameters, hyperparams]
            None means all
        """
        np = problem.npars
        if kpar is None :
            kpar = self.makeIndex( np, dval )

        elif Tools.isInstance( kpar, int ) :
            return ( problem.domain2Unit( dval, kpar ) if kpar >= 0 else
                     self.errdis.domain2Unit( dval, kpar ) )

        uval = numpy.ndarray( len( kpar ), dtype=float )

        for i,kp in enumerate( kpar ) :
            if kp >= 0 :
                uval[i] = problem.domain2Unit( dval[i], kp )
            else :
                uval[i] = self.errdis.domain2Unit( dval[i], kp )
        return uval

    def unit2Domain( self, problem, uval, kpar=None ) :
        """
        Return domain value for the selected parameter.

        Parameters
        ----------
        problem : Problem
            the problem involved
        uval : array_like
            unit value for the selected parameter
        kpar : None or array_like
            selected parameter indices, where kp is index in [parameters, hyperparams]
            None means all.
        """
        np = problem.npars

        if kpar is None :
            kpar = self.makeIndex( np, uval )
        elif Tools.isInstance( kpar, int ) :
            return ( problem.unit2Domain( uval, kpar ) if kpar >= 0 else
                     self.errdis.unit2Domain( uval, kpar ) )

        dval = numpy.ndarray( len( kpar ), dtype=float )
        for i,kp in enumerate( kpar ) :
            if kp >= 0 :
                dval[i] = problem.unit2Domain( uval[i], kp )
            else :
                dval[i] = self.errdis.unit2Domain( uval[i], kp )
        return dval

    def makeIndex( self, np, val ) :
        kpar = [k for k in range( np )]
        nh = len( val ) - np
        kpar += [-k for k in range( nh, 0, -1 )]
        return kpar

    def reportCall( self ):
        """ Store a call to engine  """
        self.report[self.NCALLS] += 1

    def reportSuccess( self ):
        """
        Add 1 to the number of succesfull steps: logL < lowLhood.
        """
        self.report[self.SUCCESS] += 1

    def reportReject( self ):
        """
        Add 1 to the number of rejected steps: logL > lowLhood.
        """
        self.report[self.REJECT] += 1

    def reportFailed( self ):
        """
        Add 1 to the number of failed steps: could not construct a step.
        """
        self.report[self.FAILED] += 1

    def reportBest( self ):
        """
        Add 1 to the number of best likelihoods found upto now.
        """
        self.report[self.BEST] += 1

    def printReport( self ) :
        print( " %10d %10d %10d %10d %10d" % (self.report[0], self.report[1],
                              self.report[2], self.report[3], self.report[4] ) )

    def successRate( self ) :
        """
        Return percentage of success.
        """
        if not hasattr( self, "save" ) :
            self.save = [0, 0]

        srate = ( 100 * ( self.report[0] - self.save[0] ) / 
                  ( self.report[0] + self.report[1] + self.report[2] - self.save[1] ) )

        self.save = [self.report[0], ( self.report[0] + self.report[1] + self.report[2] )]
        return srate

    def calculateUnitRange( self ):
        """
        Calculate the range of the present parameter values in unit values.

        For Dynamic models the range is calculated for those parameters present in all models;
        it is 1.0 for other parameters.

        """
        kmx = 0
        if not self.walkers[0].problem.model.isDynamic() :
            npmax = len( self.walkers[0].allpars )
        else :
            npmax = 0
            for k, walker in enumerate( self.walkers ) :
                if len( walker.allpars ) > npmax :
                    npmax = len( walker.allpars )
                    kmx = k

        minv = self.walkers[kmx].allpars.copy()
        maxv = self.walkers[kmx].allpars.copy()
        nval = numpy.zeros( npmax, dtype=int )

        for walker in self.walkers :
            fi = walker.fitIndex
#            print( "Eng   ", fi, minv, walker.allpars )
            minv[fi] = numpy.fmin( minv[fi], walker.allpars[fi] )
            maxv[fi] = numpy.fmax( maxv[fi], walker.allpars[fi] )
            nval[fi] += 1

        problem = self.walkers[kmx].problem
        fi = self.walkers[kmx].fitIndex

        maxv[fi] = self.domain2Unit( problem, maxv[fi], kpar=fi )
        minv[fi] = self.domain2Unit( problem, minv[fi], kpar=fi )

        self.unitRange = numpy.abs( maxv - minv )
        self.unitMin = numpy.fmin( minv, maxv )

#        print( "Eng1  ", self.unitRange )
        return

    def __str__( self ) :
        return str( "Engine" )

    def execute( self, kw, lowLhood ):
        """
        Execute the engine for diffusing the parameters

        Parameters
        ----------
        kw : walker-id
            walker to diffuse
        lowLhood : float
            low limit on the loglikelihood

        Returns
        -------
        int : number of succesfull moves

        """
        pass


class DummyPlotter( object ) :

    def __init__( self, iter=0 ) :
        self.iter = iter

    def start( self, param=None ):
        """ start the plot. """
        pass

    def point( self, param, col=None, sym=0 ):
        """
        Place a point at position param using color col and symbol sym.
        """
        pass

    def move( self, param, ptry, col=None, sym=None ):
        """
        Move parameters at position param to ptry using color col.
        """
        pass

    def stop( self, param=None, name=None ):
        """ Stop (show) the plot. """
        pass




import numpy as numpy
from astropy import units
import math
from . import Tools
from .Walker import Walker

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2008 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester


class WalkerList( list ):
    """
    WalkerList is a list of Walker.

    It is the working ensemble of NestedSampler.


    Attributes
    ----------
    logZ : float (read-only)
        Natural log of evidence
    info : float (read-only)
        The information H. The compression factor ( the ratio of the prior space
        available to the model parameters over the posterior space ) is equal to the exp( H ).
    iteration : int
        Present iteration number.

    Author       Do Kester

    """
    def __init__( self, problem, nwalkers, allpars, fitIndex ):
        """
        Constructor.

        Parameters
        ----------
        nwalkers : int
            number of walkerss created.
        problem : Problem
            to be used in the walkerss
        allpars : array_like
            parameters and hyperparams of the problem
        fitIndex : array_like
            indices of allpars to be fitted

        """
        super( WalkerList, self ).__init__( )
        self._count = 0
        self.iteration = 0
        self.logZ = 0.0
        self.info = 0.0
        allpars = numpy.asarray( allpars )
        self.addWalkers( problem, nwalkers, allpars, fitIndex )


    def addWalkers( self, problem, nWalkers, allpars, fitIndex ):
        for i in range( nWalkers ) :
            if problem.model and problem.model.isDynamic() :
                problem = problem.copy()
                fitIndex = fitIndex.copy()
            walker = Walker( self._count, problem, allpars, fitIndex )
            self.append( walker )
            self._count += 1

    # ===========================================================================
    def setWalker( self, walker, index ):
        """
        replace/append a Walker to this list

        Parameters
        ----------
        walker : Walker
            the list to take to copy from
        index : int
            the index at which to set
        """
#        print( "setWalker  ", id( self ), len( self ), walker.id, index )

        if index < len( self ) :
            self[index] = walker

        else :
            walker.id = self._count
            self._count += 1
            self.append( walker )

    def copy( self, src, des ):
        """
        Copy one item of the list onto another.

        Parameters
        ----------
        src : int
            the source item
        des : int
            the destination item

        """
        id = self[des].id
        self[des] = self[src].copy()
        self[des].id = id


    def logPlus( self, x, y ):
        """
        Return the log of sum.
        """
        return numpy.logaddexp( x, y )


    def getParameterEvolution( self, kpar=None ):
        """
        Return the evolution of one or all parameters.

        In case of dynamic models the number of parameters may vary.
        They are zero-padded. Use `getNumberOfParametersEvolution`
        to get the actual number.

        Parameters
        ----------
        kpar : int or tuple of ints
            the parameter to be selected. Default: all

        """
        pe = []
        for walker in self :
            pe += [walker.parameters]
        if kpar is None :
            return numpy.asarray( pe )
        else :
            return numpy.asarray( pe )[:,kpar]

    def getScaleEvolution( self ):
        """ Return the evolution of the scale.  """
        pe = [walker.hypars for walker in self]
        return numpy.asarray( pe )

    def getLogLikelihoodEvolution( self ):
        """ Return the evolution of the log( Likelihood ).  """
        pe = [walker.logL for walker in self]
        return numpy.asarray( pe )

    def getLowLogL( self ):
        """
        Return the lowest value of logL in the walkerlist, plus its index.
        """
        low = self[0].logL
        klo = 0
        k = 0
        for walker in self :
            if walker.logL < low :
                klo = k
                low = walker.logL
            k += 1
        return ( low, klo )


